from __future__ import annotations

import random
from functools import cmp_to_key
from typing import TypeVar, Optional, Callable, Iterable, List, Type

from kotcollections.kot_list import KotList

T = TypeVar('T')


class KotMutableList(KotList[T]):
    def __init__(self, elements: Optional[Iterable[T]] = None):
        super().__init__(elements)

    @classmethod
    def __class_getitem__(cls, element_type: Type[T]) -> Type['KotMutableList[T]']:
        """Enable KotMutableList[Type]() syntax for type specification.
        
        Example:
            animals = KotMutableList[Animal]()
            animals.add(Dog("Buddy"))
            animals.add(Cat("Whiskers"))
        """
        class TypedKotMutableList(cls):
            def __init__(self, elements=None):
                # Only set element type if it's an actual type, not a type variable
                if isinstance(element_type, type):
                    self._element_type = element_type
                else:
                    self._element_type = None
                self._elements = []
                # Now process elements with the correct type set
                if elements is not None:
                    for elem in elements:
                        self._check_type(elem)
                        self._elements.append(elem)
        
        # Set a meaningful name for debugging
        TypedKotMutableList.__name__ = f"{cls.__name__}[{element_type.__name__}]"
        TypedKotMutableList.__qualname__ = f"{cls.__qualname__}[{element_type.__name__}]"
        
        return TypedKotMutableList

    @classmethod
    def of_type(cls, element_type: Type[T], elements: Optional[Iterable[T]] = None) -> 'KotMutableList[T]':
        """Create a KotMutableList with a specific element type.
        
        This is useful when you want to create a mutable list of a parent type
        but only have instances of child types.
        
        Args:
            element_type: The type of elements this list will contain
            elements: Optional initial elements
            
        Returns:
            A new KotMutableList instance with the specified element type
            
        Example:
            animals = KotMutableList.of_type(Animal, [Dog("Buddy"), Cat("Whiskers")])
            # or empty list
            animals = KotMutableList.of_type(Animal)
            animals.add(Dog("Max"))
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[element_type]
        return typed_class(elements)

    def __setitem__(self, index: int, value: T) -> None:
        self.set(index, value)

    def __delitem__(self, index: int) -> None:
        self.remove_at(index)

    def add(self, element: T) -> bool:
        self._check_type(element)
        self._elements.append(element)
        return True

    def add_at(self, index: int, element: T) -> None:
        if not 0 <= index <= self.size:
            raise IndexError(f"Index {index} out of bounds for insertion")
        self._check_type(element)
        self._elements.insert(index, element)

    def add_all(self, elements: Iterable[T]) -> bool:
        elements_list = list(elements)
        if elements_list:
            for element in elements_list:
                self._check_type(element)
            self._elements.extend(elements_list)
            return True
        return False

    def add_all_at(self, index: int, elements: Iterable[T]) -> bool:
        if not 0 <= index <= self.size:
            raise IndexError(f"Index {index} out of bounds for insertion")
        elements_list = list(elements)
        if elements_list:
            for element in elements_list:
                self._check_type(element)
            for i, element in enumerate(elements_list):
                self._elements.insert(index + i, element)
            return True
        return False

    def set(self, index: int, element: T) -> T:
        if not 0 <= index < self.size:
            raise IndexError(f"Index {index} out of bounds for list of size {self.size}")
        self._check_type(element)
        old_element = self._elements[index]
        self._elements[index] = element
        return old_element

    def remove_at(self, index: int) -> T:
        if not 0 <= index < self.size:
            raise IndexError(f"Index {index} out of bounds for list of size {self.size}")
        return self._elements.pop(index)

    def remove(self, element: T) -> bool:
        try:
            self._elements.remove(element)
            return True
        except ValueError:
            return False

    def remove_all(self, elements: Iterable[T]) -> bool:
        elements_set = set(elements)
        initial_size = self.size
        self._elements = [e for e in self._elements if e not in elements_set]
        return self.size < initial_size

    def remove_first(self) -> T:
        """Removes the first element from this mutable list."""
        if self.is_empty():
            raise IndexError("List is empty")
        return self._elements.pop(0)

    def remove_last(self) -> T:
        """Removes the last element from this mutable list."""
        if self.is_empty():
            raise IndexError("List is empty")
        return self._elements.pop()

    def remove_first_or_null(self) -> Optional[T]:
        """Removes the first element from this mutable list and returns it, or null if the list is empty."""
        if self.is_empty():
            return None
        return self._elements.pop(0)

    def remove_first_or_none(self) -> Optional[T]:
        """Alias for remove_first_or_null() - more Pythonic naming."""
        return self.remove_first_or_null()

    def remove_last_or_null(self) -> Optional[T]:
        """Removes the last element from this mutable list and returns it, or null if the list is empty."""
        if self.is_empty():
            return None
        return self._elements.pop()

    def remove_last_or_none(self) -> Optional[T]:
        """Alias for remove_last_or_null() - more Pythonic naming."""
        return self.remove_last_or_null()

    def retain_all(self, elements: Iterable[T]) -> bool:
        elements_set = set(elements)
        initial_size = self.size
        self._elements = [e for e in self._elements if e in elements_set]
        return self.size < initial_size

    def clear(self) -> None:
        self._elements.clear()

    def sort(self, key: Optional[Callable[[T], any]] = None, reverse: bool = False) -> None:
        self._elements.sort(key=key, reverse=reverse)

    def sort_descending(self) -> None:
        self._elements.sort(reverse=True)

    def sort_by(self, selector: Callable[[T], any]) -> None:
        self._elements.sort(key=selector)

    def sort_by_descending(self, selector: Callable[[T], any]) -> None:
        self._elements.sort(key=selector, reverse=True)

    def sort_with(self, comparator: Callable[[T, T], int]) -> None:
        """Sorts elements in the list in-place according to the specified comparator."""
        self._elements.sort(key=cmp_to_key(comparator))

    def reverse(self) -> None:
        self._elements.reverse()

    def shuffle(self, random_instance: Optional[random.Random] = None) -> None:
        if random_instance:
            random_instance.shuffle(self._elements)
        else:
            random.shuffle(self._elements)

    def fill(self, value: T) -> None:
        for i in range(self.size):
            self._elements[i] = value

    def as_reversed(self) -> 'KotMutableList[T]':
        class KotReversedMutableList(KotMutableList[T]):
            def __init__(self, original: KotMutableList[T]):
                self._original = original
                super().__init__()

            @property
            def _elements(self) -> List[T]:
                return list(reversed(self._original._elements))

            @_elements.setter
            def _elements(self, value: List[T]) -> None:
                pass

            def __getitem__(self, index: int) -> T:
                return self._original._elements[self._original.size - 1 - index]

            def __setitem__(self, index: int, value: T) -> None:
                self._original._elements[self._original.size - 1 - index] = value

            def __len__(self) -> int:
                return self._original.size

        return KotReversedMutableList(self)



from __future__ import annotations

import bisect
import random as _random
from collections.abc import Iterable
from functools import reduce, cmp_to_key
from typing import TypeVar, Generic, Callable, Optional, List, Tuple, Iterator, Any, Dict, Union, TYPE_CHECKING, Set, Type

T = TypeVar('T')
R = TypeVar('R')
K = TypeVar('K')
V = TypeVar('V')

if TYPE_CHECKING:
    from kotcollections.kot_map import KotMap
    from kotcollections.kot_mutable_list import KotMutableList
    from kotcollections.kot_set import KotSet
    from kotcollections.kot_mutable_set import KotMutableSet



class KotList(Generic[T]):
    def __init__(self, elements: Optional[Iterable[T]] = None):
        self._element_type: Optional[type] = None
        if elements is None:
            self._elements: List[T] = []
        else:
            elements_list = list(elements)
            if elements_list:
                # Set the element type based on the first element
                first_elem = elements_list[0]
                if isinstance(first_elem, KotList):
                    self._element_type = KotList
                else:
                    self._element_type = type(first_elem)

                # Check all elements have the same type
                for elem in elements_list:
                    self._check_type(elem)

            self._elements = elements_list

    @classmethod
    def __class_getitem__(cls, element_type: Type[T]) -> Type['KotList[T]']:
        """Enable KotList[Type]() syntax for type specification.
        
        Example:
            animals = KotList[Animal]()
            animals_mutable = animals.to_kot_mutable_list()
            animals_mutable.add(Dog("Buddy"))
            animals_mutable.add(Cat("Whiskers"))
        """
        class TypedKotList(cls):
            def __init__(self, elements=None):
                # Only set element type if it's an actual type, not a type variable
                if isinstance(element_type, type):
                    self._element_type = element_type
                else:
                    self._element_type = None
                self._elements = []
                # Now process elements with the correct type set
                if elements is not None:
                    for elem in elements:
                        self._check_type(elem)
                        self._elements.append(elem)
        
        # Set a meaningful name for debugging
        TypedKotList.__name__ = f"{cls.__name__}[{element_type.__name__}]"
        TypedKotList.__qualname__ = f"{cls.__qualname__}[{element_type.__name__}]"
        
        return TypedKotList

    @classmethod
    def of_type(cls, element_type: Type[T], elements: Optional[Iterable[T]] = None) -> 'KotList[T]':
        """Create a KotList with a specific element type.
        
        This is useful when you want to create a list of a parent type
        but only have instances of child types.
        
        Args:
            element_type: The type of elements this list will contain
            elements: Optional initial elements
            
        Returns:
            A new KotList instance with the specified element type
            
        Example:
            animals = KotList.of_type(Animal, [Dog("Buddy"), Cat("Whiskers")])
            # or empty list
            animals = KotList.of_type(Animal)
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[element_type]
        return typed_class(elements)

    def _check_type(self, element: Any) -> None:
        """Check if the element has the correct type for this list."""
        # Skip type checking if element_type is a type variable or not a real type
        if self._element_type is not None and not isinstance(self._element_type, type):
            return
            
        if self._element_type is None:
            # First element sets the type
            if isinstance(element, KotList):
                self._element_type = KotList
            else:
                self._element_type = type(element)
        else:
            # Type check: allow T type or KotList type
            if isinstance(element, KotList):
                if self._element_type == KotList:
                    pass  # KotList type is allowed
                # Special handling for __class_getitem__ types (e.g., KotList[Task])
                elif (hasattr(self._element_type, '__base__') and 
                      hasattr(self._element_type, '__name__') and 
                      '[' in self._element_type.__name__ and
                      isinstance(element, self._element_type.__base__)):
                    # Check if the element has matching element type for KotList/KotSet/KotMap types
                    if hasattr(element, '_element_type') and hasattr(self._element_type, '__new__'):
                        # Extract expected element type from the __class_getitem__ type
                        # This is a more strict check for collection types
                        pass
                    else:
                        pass
                else:
                    type_name = getattr(self._element_type, '__name__', str(self._element_type))
                    raise TypeError(f"Cannot add KotList to KotList[{type_name}]")
            elif not isinstance(element, self._element_type):
                # Check if element is an instance of the expected type
                if isinstance(element, self._element_type):
                    pass  # Direct instance check passed (won't reach here, but for clarity)
                # Special handling for __class_getitem__ types
                elif (hasattr(self._element_type, '__base__') and 
                      hasattr(self._element_type, '__name__') and 
                      '[' in self._element_type.__name__ and
                      isinstance(element, self._element_type.__base__)):
                    # Allow instances of base class for __class_getitem__ types
                    pass
                else:
                    raise TypeError(
                        f"Cannot add element of type '{type(element).__name__}' to KotList[{self._element_type.__name__}]"
                    )

    def __repr__(self) -> str:
        return f"KotList({self._elements})"

    def __str__(self) -> str:
        return str(self._elements)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, KotList):
            return False
        return self._elements == other._elements

    def __hash__(self) -> int:
        return hash(tuple(self._elements))

    def __iter__(self) -> Iterator[T]:
        return iter(self._elements)

    def __getitem__(self, index: int) -> T:
        return self.get(index)

    def __len__(self) -> int:
        return self.size

    @property
    def size(self) -> int:
        return len(self._elements)

    @property
    def indices(self) -> range:
        return range(self.size)

    @property
    def last_index(self) -> int:
        return self.size - 1 if self.size > 0 else -1

    def is_empty(self) -> bool:
        return self.size == 0

    def is_not_empty(self) -> bool:
        return self.size > 0

    def get(self, index: int) -> T:
        if not 0 <= index < self.size:
            raise IndexError(f"Index {index} out of bounds for list of size {self.size}")
        return self._elements[index]

    def get_or_null(self, index: int) -> Optional[T]:
        return self._elements[index] if 0 <= index < self.size else None

    def get_or_none(self, index: int) -> Optional[T]:
        """Alias for get_or_null() - more Pythonic naming."""
        return self.get_or_null(index)

    def get_or_else(self, index: int, default_value: Callable[[int], T]) -> T:
        return self._elements[index] if 0 <= index < self.size else default_value(index)

    def first(self) -> T:
        if self.is_empty():
            raise IndexError("List is empty")
        return self._elements[0]

    def first_predicate(self, predicate: Callable[[T], bool]) -> T:
        for element in self._elements:
            if predicate(element):
                return element
        raise ValueError("No element matching predicate found")

    def first_or_null(self) -> Optional[T]:
        return self._elements[0] if self.is_not_empty() else None

    def first_or_none(self) -> Optional[T]:
        """Alias for first_or_null() - more Pythonic naming."""
        return self.first_or_null()

    def first_or_null_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        for element in self._elements:
            if predicate(element):
                return element
        return None

    def first_or_none_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Alias for first_or_null_predicate() - more Pythonic naming."""
        return self.first_or_null_predicate(predicate)

    def last(self) -> T:
        if self.is_empty():
            raise IndexError("List is empty")
        return self._elements[-1]

    def last_predicate(self, predicate: Callable[[T], bool]) -> T:
        for element in reversed(self._elements):
            if predicate(element):
                return element
        raise ValueError("No element matching predicate found")

    def last_or_null(self) -> Optional[T]:
        return self._elements[-1] if self.is_not_empty() else None

    def last_or_none(self) -> Optional[T]:
        """Alias for last_or_null() - more Pythonic naming."""
        return self.last_or_null()

    def last_or_null_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        for element in reversed(self._elements):
            if predicate(element):
                return element
        return None

    def last_or_none_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Alias for last_or_null_predicate() - more Pythonic naming."""
        return self.last_or_null_predicate(predicate)

    def element_at(self, index: int) -> T:
        return self.get(index)

    def element_at_or_else(self, index: int, default_value: Callable[[int], T]) -> T:
        return self.get_or_else(index, default_value)

    def element_at_or_null(self, index: int) -> Optional[T]:
        return self.get_or_null(index)

    def element_at_or_none(self, index: int) -> Optional[T]:
        """Alias for element_at_or_null() - more Pythonic naming."""
        return self.element_at_or_null(index)

    def contains(self, element: T) -> bool:
        return element in self._elements

    def contains_all(self, elements: Iterable[T]) -> bool:
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(elements, KotSet):
            elements_set = set(elements)
        elif isinstance(elements, KotMap):
            elements_set = set(elements.values)
        else:
            elements_set = set(elements)
        return all(elem in self._elements for elem in elements_set)

    def index_of(self, element: T) -> int:
        try:
            return self._elements.index(element)
        except ValueError:
            return -1

    def last_index_of(self, element: T) -> int:
        for i in range(len(self._elements) - 1, -1, -1):
            if self._elements[i] == element:
                return i
        return -1

    def index_of_first(self, predicate: Callable[[T], bool]) -> int:
        for i, element in enumerate(self._elements):
            if predicate(element):
                return i
        return -1

    def index_of_last(self, predicate: Callable[[T], bool]) -> int:
        for i in range(len(self._elements) - 1, -1, -1):
            if predicate(self._elements[i]):
                return i
        return -1

    def binary_search(self, element: T, comparator: Optional[Callable[[T, T], int]] = None) -> int:
        if comparator is None:
            index = bisect.bisect_left(self._elements, element)
            if index < len(self._elements) and self._elements[index] == element:
                return index
            return -(index + 1)
        else:
            left, right = 0, len(self._elements) - 1
            while left <= right:
                mid = (left + right) // 2
                cmp = comparator(self._elements[mid], element)
                if cmp < 0:
                    left = mid + 1
                elif cmp > 0:
                    right = mid - 1
                else:
                    return mid
            return -(left + 1)

    def binary_search_by(
        self,
        key: K,
        selector: Callable[[T], K],
        comparator: Optional[Callable[[K, K], int]] = None
    ) -> int:
        """Searches this list or its range for an element having the key returned by the specified selector function equal to the provided key value using the binary search algorithm."""
        if comparator is None:
            # Create a list of (key, index) pairs for binary search
            keys = [(selector(elem), i) for i, elem in enumerate(self._elements)]
            keys.sort(key=lambda x: x[0])
            index = bisect.bisect_left([k[0] for k in keys], key)
            if index < len(keys) and keys[index][0] == key:
                return keys[index][1]
            return -(index + 1)
        else:
            # Custom comparator case
            left, right = 0, len(self._elements) - 1
            while left <= right:
                mid = (left + right) // 2
                mid_key = selector(self._elements[mid])
                cmp = comparator(mid_key, key)
                if cmp < 0:
                    left = mid + 1
                elif cmp > 0:
                    right = mid - 1
                else:
                    return mid
            return -(left + 1)

    def map(self, transform: Callable[[T], R]) -> 'KotList[R]':
        return KotList([transform(element) for element in self._elements])

    def map_indexed(self, transform: Callable[[int, T], R]) -> 'KotList[R]':
        return KotList([transform(i, element) for i, element in enumerate(self._elements)])

    def map_not_null(self, transform: Callable[[T], Optional[R]]) -> 'KotList[R]':
        result = []
        for element in self._elements:
            transformed = transform(element)
            if transformed is not None:
                result.append(transformed)
        return KotList(result)

    def map_not_none(self, transform: Callable[[T], Optional[R]]) -> 'KotList[R]':
        """Alias for map_not_null() - more Pythonic naming."""
        return self.map_not_null(transform)

    def flat_map(self, transform: Callable[[T], Iterable[R]]) -> 'KotList[R]':
        result = []
        for element in self._elements:
            result.extend(transform(element))
        return KotList(result)

    def flatten(self) -> 'KotList[Any]':
        result = []
        for element in self._elements:
            if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
                result.extend(element)
            else:
                result.append(element)
        return KotList(result)

    def associate_with(self, value_selector: Callable[[T], V]) -> 'KotMap[T, V]':
        from kotcollections.kot_map import KotMap
        return KotMap({element: value_selector(element) for element in self._elements})

    def associate_by(self, key_selector: Callable[[T], K]) -> 'KotMap[K, T]':
        from kotcollections.kot_map import KotMap
        result = {}
        for element in self._elements:
            result[key_selector(element)] = element
        return KotMap(result)

    def associate_by_with_value(
        self, key_selector: Callable[[T], K],
        value_transform: Callable[[T], V]
    ) -> 'KotMap[K, V]':
        from kotcollections.kot_map import KotMap
        result = {}
        for element in self._elements:
            result[key_selector(element)] = value_transform(element)
        return KotMap(result)

    def filter(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        return KotList([element for element in self._elements if predicate(element)])

    def filter_indexed(self, predicate: Callable[[int, T], bool]) -> 'KotList[T]':
        return KotList([element for i, element in enumerate(self._elements) if predicate(i, element)])

    def filter_not(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        return KotList([element for element in self._elements if not predicate(element)])

    def filter_not_null(self) -> 'KotList[T]':
        return KotList([element for element in self._elements if element is not None])

    def filter_not_none(self) -> 'KotList[T]':
        """Alias for filter_not_null() - more Pythonic naming."""
        return self.filter_not_null()

    def filter_is_instance(self, klass: type) -> 'KotList[Any]':
        return KotList([element for element in self._elements if isinstance(element, klass)])

    def partition(self, predicate: Callable[[T], bool]) -> Tuple['KotList[T]', 'KotList[T]']:
        matching = []
        non_matching = []
        for element in self._elements:
            if predicate(element):
                matching.append(element)
            else:
                non_matching.append(element)
        return KotList(matching), KotList(non_matching)

    def any(self, predicate: Optional[Callable[[T], bool]] = None) -> bool:
        if predicate is None:
            return self.is_not_empty()
        return any(predicate(element) for element in self._elements)

    def all(self, predicate: Callable[[T], bool]) -> bool:
        return all(predicate(element) for element in self._elements)

    def none(self, predicate: Optional[Callable[[T], bool]] = None) -> bool:
        if predicate is None:
            return self.is_empty()
        return not any(predicate(element) for element in self._elements)

    def count(self, predicate: Optional[Callable[[T], bool]] = None) -> int:
        if predicate is None:
            return self.size
        return sum(1 for element in self._elements if predicate(element))

    def sum_of(self, selector: Callable[[T], Union[int, float]]) -> Union[int, float]:
        return sum(selector(element) for element in self._elements)

    def max_or_null(self) -> Optional[T]:
        return max(self._elements) if self.is_not_empty() else None

    def max_or_none(self) -> Optional[T]:
        """Alias for max_or_null() - more Pythonic naming."""
        return self.max_or_null()

    def min_or_null(self) -> Optional[T]:
        return min(self._elements) if self.is_not_empty() else None

    def min_or_none(self) -> Optional[T]:
        """Alias for min_or_null() - more Pythonic naming."""
        return self.min_or_null()

    def max_by_or_null(self, selector: Callable[[T], Any]) -> Optional[T]:
        if self.is_empty():
            return None
        return max(self._elements, key=selector)

    def max_by_or_none(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Alias for max_by_or_null() - more Pythonic naming."""
        return self.max_by_or_null(selector)

    def min_by_or_null(self, selector: Callable[[T], Any]) -> Optional[T]:
        if self.is_empty():
            return None
        return min(self._elements, key=selector)

    def min_by_or_none(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Alias for min_by_or_null() - more Pythonic naming."""
        return self.min_by_or_null(selector)

    def average(self) -> float:
        if self.is_empty():
            raise ValueError("Cannot compute average of empty list")
        return sum(self._elements) / self.size

    def sorted(self, key: Optional[Callable[[T], Any]] = None, reverse: bool = False) -> 'KotList[T]':
        return KotList(sorted(self._elements, key=key, reverse=reverse))

    def sorted_descending(self) -> 'KotList[T]':
        return KotList(sorted(self._elements, reverse=True))

    def sorted_by(self, selector: Callable[[T], Any]) -> 'KotList[T]':
        return KotList(sorted(self._elements, key=selector))

    def sorted_by_descending(self, selector: Callable[[T], Any]) -> 'KotList[T]':
        return KotList(sorted(self._elements, key=selector, reverse=True))

    def sorted_with(self, comparator: Callable[[T, T], int]) -> 'KotList[T]':
        """Returns a list of all elements sorted according to the specified comparator."""
        return KotList(sorted(self._elements, key=cmp_to_key(comparator)))

    def reversed(self) -> 'KotList[T]':
        return KotList(reversed(self._elements))

    def shuffled(self, random_instance: Optional[_random.Random] = None) -> 'KotList[T]':
        elements_copy = self._elements.copy()
        if random_instance:
            random_instance.shuffle(elements_copy)
        else:
            _random.shuffle(elements_copy)
        return KotList(elements_copy)

    def group_by(self, key_selector: Callable[[T], K]) -> 'KotMap[K, KotList[T]]':
        from kotcollections.kot_map import KotMap
        result: Dict[K, List[T]] = {}
        for element in self._elements:
            key = key_selector(element)
            if key not in result:
                result[key] = []
            result[key].append(element)
        return KotMap({k: KotList(v) for k, v in result.items()})

    def group_by_with_value(
        self, key_selector: Callable[[T], K],
        value_transform: Callable[[T], V]
    ) -> 'KotMap[K, KotList[V]]':
        from kotcollections.kot_map import KotMap
        result: Dict[K, List[V]] = {}
        for element in self._elements:
            key = key_selector(element)
            if key not in result:
                result[key] = []
            result[key].append(value_transform(element))
        return KotMap({k: KotList(v) for k, v in result.items()})

    def chunked(self, size: int) -> 'KotList[KotList[T]]':
        if size <= 0:
            raise ValueError("Size must be positive")
        chunks = []
        for i in range(0, len(self._elements), size):
            chunks.append(KotList(self._elements[i:i + size]))
        return KotList(chunks)

    def chunked_transform(self, size: int, transform: Callable[['KotList[T]'], R]) -> 'KotList[R]':
        if size <= 0:
            raise ValueError("Size must be positive")
        result = []
        for i in range(0, len(self._elements), size):
            chunk = KotList(self._elements[i:i + size])
            result.append(transform(chunk))
        return KotList(result)

    def windowed(self, size: int, step: int = 1, partial_windows: bool = False) -> 'KotList[KotList[T]]':
        if size <= 0 or step <= 0:
            raise ValueError("Size and step must be positive")
        windows = []
        for i in range(0, len(self._elements), step):
            window = self._elements[i:i + size]
            if len(window) == size or (partial_windows and window):
                windows.append(KotList(window))
            elif not partial_windows and len(window) < size:
                break
        return KotList(windows)

    def distinct(self) -> 'KotList[T]':
        seen = set()
        result = []
        for element in self._elements:
            if element not in seen:
                seen.add(element)
                result.append(element)
        return KotList(result)

    def distinct_by(self, selector: Callable[[T], K]) -> 'KotList[T]':
        seen = set()
        result = []
        for element in self._elements:
            key = selector(element)
            if key not in seen:
                seen.add(key)
                result.append(element)
        return KotList(result)

    def intersect(self, other: Iterable[T]) -> 'KotList[T]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(other, KotSet):
            other_set = set(other)
        elif isinstance(other, KotMap):
            other_set = set(other.values)
        else:
            other_set = set(other)
        return KotList([element for element in self._elements if element in other_set])

    def union(self, other: Iterable[T]) -> 'KotList[T]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        result = list(self._elements)
        seen = set(self._elements)
        
        if isinstance(other, KotSet):
            iter_other = other
        elif isinstance(other, KotMap):
            iter_other = other.values
        else:
            iter_other = other
            
        for element in iter_other:
            if element not in seen:
                result.append(element)
                seen.add(element)
        return KotList(result)

    def plus(self, element: Union[T, Iterable[T]]) -> 'KotList[T]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
            if isinstance(element, KotSet):
                return KotList(self._elements + list(element))
            elif isinstance(element, KotMap):
                return KotList(self._elements + list(element.values))
            else:
                return KotList(self._elements + list(element))
        else:
            return KotList(self._elements + [element])

    def minus(self, element: Union[T, Iterable[T]]) -> 'KotList[T]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
            if isinstance(element, KotSet):
                to_remove = set(element)
            elif isinstance(element, KotMap):
                to_remove = set(element.values)
            else:
                to_remove = set(element)
            return KotList([e for e in self._elements if e not in to_remove])
        else:
            result = self._elements.copy()
            if element in result:
                result.remove(element)
            return KotList(result)

    def sub_list(self, from_index: int, to_index: int) -> 'KotList[T]':
        return KotList(self._elements[from_index:to_index])

    def zip(self, other: Iterable[R]) -> 'KotList[Tuple[T, R]]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(other, KotSet):
            return KotList(list(zip(self._elements, other)))
        elif isinstance(other, KotMap):
            return KotList(list(zip(self._elements, other.values)))
        else:
            return KotList(list(zip(self._elements, other)))

    def zip_transform(self, other: Iterable[R], transform: Callable[[T, R], V]) -> 'KotList[V]':
        # Support KotSet and KotMap explicitly
        from kotcollections.kot_set import KotSet
        from kotcollections.kot_map import KotMap
        
        if isinstance(other, KotSet):
            iter_other = other
        elif isinstance(other, KotMap):
            iter_other = other.values
        else:
            iter_other = other
        
        return KotList([transform(a, b) for a, b in zip(self._elements, iter_other)])

    def unzip(self) -> Tuple['KotList[Any]', 'KotList[Any]']:
        if self.is_empty():
            return KotList(), KotList()
        first_elements = []
        second_elements = []
        for pair in self._elements:
            first_elements.append(pair[0])
            second_elements.append(pair[1])
        return KotList(first_elements), KotList(second_elements)

    def fold(self, initial: R, operation: Callable[[R, T], R]) -> R:
        result = initial
        for element in self._elements:
            result = operation(result, element)
        return result

    def reduce(self, operation: Callable[[T, T], T]) -> T:
        if self.is_empty():
            raise ValueError("Cannot reduce empty list")
        return reduce(operation, self._elements)

    def scan(self, initial: R, operation: Callable[[R, T], R]) -> 'KotList[R]':
        result = [initial]
        acc = initial
        for element in self._elements:
            acc = operation(acc, element)
            result.append(acc)
        return KotList(result)

    def for_each(self, action: Callable[[T], None]) -> None:
        for element in self._elements:
            action(element)

    def for_each_indexed(self, action: Callable[[int, T], None]) -> None:
        for i, element in enumerate(self._elements):
            action(i, element)

    def on_each(self, action: Callable[[T], None]) -> 'KotList[T]':
        for element in self._elements:
            action(element)
        return self

    def to_list(self) -> List[T]:
        return self._elements.copy()

    def to_set(self) -> Set[T]:
        return set(self._elements.copy())

    def to_kot_list(self) -> 'KotList[T]':
        # Preserve type information when converting
        if self._element_type is not None:
            return KotList.of_type(self._element_type, self._elements.copy())
        else:
            return KotList(self._elements.copy())

    def to_kot_mutable_list(self) -> 'KotMutableList[T]':
        from kotcollections.kot_mutable_list import KotMutableList
        # Preserve type information when converting
        if self._element_type is not None:
            mutable_list = KotMutableList.of_type(self._element_type, self._elements.copy())
        else:
            mutable_list = KotMutableList(self._elements.copy())
        return mutable_list

    def to_kot_set(self) -> 'KotSet[T]':
        from kotcollections.kot_set import KotSet
        # Preserve type information when converting
        if self._element_type is not None:
            return KotSet.of_type(self._element_type, self._elements.copy())
        else:
            return KotSet(self._elements.copy())

    def to_kot_mutable_set(self) -> 'KotMutableSet[T]':
        from kotcollections.kot_mutable_set import KotMutableSet
        # Preserve type information when converting
        if self._element_type is not None:
            return KotMutableSet.of_type(self._element_type, self._elements.copy())
        else:
            return KotMutableSet(self._elements.copy())

    def join_to_string(
        self, separator: str = ", ", prefix: str = "", postfix: str = "",
        limit: int = -1, truncated: str = "...",
        transform: Optional[Callable[[T], str]] = None
    ) -> str:
        if transform is None:
            transform = str

        result = prefix
        count = 0

        for i, element in enumerate(self._elements):
            if limit >= 0 and count >= limit:
                result += truncated
                break

            if i > 0:
                result += separator

            result += transform(element)
            count += 1

        result += postfix
        return result

    # Element retrieval methods
    def component1(self) -> T:
        """Returns the first element (for destructuring declarations)."""
        return self.get(0)

    def component2(self) -> T:
        """Returns the second element (for destructuring declarations)."""
        return self.get(1)

    def component3(self) -> T:
        """Returns the third element (for destructuring declarations)."""
        return self.get(2)

    def component4(self) -> T:
        """Returns the fourth element (for destructuring declarations)."""
        return self.get(3)

    def component5(self) -> T:
        """Returns the fifth element (for destructuring declarations)."""
        return self.get(4)

    def single(self) -> T:
        """Returns the single element, or throws an exception if the list is empty or has more than one element."""
        if self.size == 0:
            raise ValueError("List is empty")
        if self.size > 1:
            raise ValueError("List has more than one element")
        return self._elements[0]

    def single_or_null(self) -> Optional[T]:
        """Returns the single element, or null if the list is empty or has more than one element."""
        return self._elements[0] if self.size == 1 else None

    def single_or_none(self) -> Optional[T]:
        """Alias for single_or_null() - more Pythonic naming."""
        return self.single_or_null()

    def single_predicate(self, predicate: Callable[[T], bool]) -> T:
        """Returns the single element matching the given predicate, or throws exception if there is no or more than one matching element."""
        found = None
        found_multiple = False

        for element in self._elements:
            if predicate(element):
                if found is not None:
                    found_multiple = True
                    break
                found = element

        if found is None:
            raise ValueError("No element matching predicate found")
        if found_multiple:
            raise ValueError("More than one element matching predicate found")

        return found

    def single_or_null_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the single element matching the given predicate, or null if element was not found or more than one element was found."""
        found = None

        for element in self._elements:
            if predicate(element):
                if found is not None:
                    return None
                found = element

        return found

    def single_or_none_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Alias for single_or_null_predicate() - more Pythonic naming."""
        return self.single_or_null_predicate(predicate)

    def random(self, random_instance: Optional[_random.Random] = None) -> T:
        """Returns a random element from this list."""
        if self.is_empty():
            raise IndexError("List is empty")
        if random_instance:
            return random_instance.choice(self._elements)
        return _random.choice(self._elements)

    def random_or_null(self, random_instance: Optional[_random.Random] = None) -> Optional[T]:
        """Returns a random element from this list, or null if this list is empty."""
        if self.is_empty():
            return None
        if random_instance:
            return random_instance.choice(self._elements)
        return _random.choice(self._elements)

    def random_or_none(self, random_instance: Optional[_random.Random] = None) -> Optional[T]:
        """Alias for random_or_null() - more Pythonic naming."""
        return self.random_or_null(random_instance)

    # Sublist retrieval methods
    def slice(self, indices: Iterable[int]) -> 'KotList[T]':
        """Returns a list containing elements at specified indices."""
        result = []
        for index in indices:
            if 0 <= index < self.size:
                result.append(self._elements[index])
            else:
                raise IndexError(f"Index {index} out of bounds for list of size {self.size}")
        return KotList(result)

    def take(self, n: int) -> 'KotList[T]':
        """Returns a list containing first n elements."""
        if n < 0:
            raise ValueError("Requested element count is less than zero")
        return KotList(self._elements[:n])

    def take_last(self, n: int) -> 'KotList[T]':
        """Returns a list containing last n elements."""
        if n < 0:
            raise ValueError("Requested element count is less than zero")
        if n == 0:
            return KotList()
        return KotList(self._elements[-n:])

    def take_while(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        """Returns a list containing first elements satisfying the given predicate."""
        result = []
        for element in self._elements:
            if predicate(element):
                result.append(element)
            else:
                break
        return KotList(result)

    def take_last_while(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        """Returns a list containing last elements satisfying the given predicate."""
        result = []
        for element in reversed(self._elements):
            if predicate(element):
                result.append(element)
            else:
                break
        return KotList(reversed(result))

    def drop(self, n: int) -> 'KotList[T]':
        """Returns a list containing all elements except first n elements."""
        if n < 0:
            raise ValueError("Requested element count is less than zero")
        return KotList(self._elements[n:])

    def drop_last(self, n: int) -> 'KotList[T]':
        """Returns a list containing all elements except last n elements."""
        if n < 0:
            raise ValueError("Requested element count is less than zero")
        if n == 0:
            return KotList(self._elements)
        return KotList(self._elements[:-n])

    def drop_while(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        """Returns a list containing all elements except first elements that satisfy the given predicate."""
        index = 0
        for i, element in enumerate(self._elements):
            if not predicate(element):
                index = i
                break
        else:
            return KotList()
        return KotList(self._elements[index:])

    def drop_last_while(self, predicate: Callable[[T], bool]) -> 'KotList[T]':
        """Returns a list containing all elements except last elements that satisfy the given predicate."""
        index = len(self._elements)
        for i in range(len(self._elements) - 1, -1, -1):
            if not predicate(self._elements[i]):
                index = i + 1
                break
        else:
            return KotList()
        return KotList(self._elements[:index])

    # Transformation methods
    def map_indexed_not_null(self, transform: Callable[[int, T], Optional[R]]) -> 'KotList[R]':
        """Returns a list containing only the non-null results of applying the given transform function to each element and its index."""
        result = []
        for i, element in enumerate(self._elements):
            transformed = transform(i, element)
            if transformed is not None:
                result.append(transformed)
        return KotList(result)

    def map_indexed_not_none(self, transform: Callable[[int, T], Optional[R]]) -> 'KotList[R]':
        """Alias for map_indexed_not_null() - more Pythonic naming."""
        return self.map_indexed_not_null(transform)

    def flat_map_indexed(self, transform: Callable[[int, T], Iterable[R]]) -> 'KotList[R]':
        """Returns a single list of all elements yielded from results of transform function being invoked on each element and its index."""
        result = []
        for i, element in enumerate(self._elements):
            result.extend(transform(i, element))
        return KotList(result)

    def zip_with_next(self) -> 'KotList[Tuple[T, T]]':
        """Returns a list of pairs of each two adjacent elements in this list."""
        if self.size < 2:
            return KotList()
        result = []
        for i in range(self.size - 1):
            result.append((self._elements[i], self._elements[i + 1]))
        return KotList(result)

    def zip_with_next_transform(self, transform: Callable[[T, T], R]) -> 'KotList[R]':
        """Returns a list containing the results of applying the given transform function to each pair of two adjacent elements."""
        if self.size < 2:
            return KotList()
        result = []
        for i in range(self.size - 1):
            result.append(transform(self._elements[i], self._elements[i + 1]))
        return KotList(result)

    # Search methods
    def find(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the first element matching the given predicate, or null if no such element was found."""
        return self.first_or_null_predicate(predicate)

    def find_last(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the last element matching the given predicate, or null if no such element was found."""
        return self.last_or_null_predicate(predicate)

    def first_not_null_of(self, transform: Callable[[T], Optional[R]]) -> R:
        """Returns the first non-null value produced by transform function or throws NoSuchElementException."""
        for element in self._elements:
            result = transform(element)
            if result is not None:
                return result
        raise ValueError("No non-null value found")

    def first_not_null_of_or_null(self, transform: Callable[[T], Optional[R]]) -> Optional[R]:
        """Returns the first non-null value produced by transform function or null."""
        for element in self._elements:
            result = transform(element)
            if result is not None:
                return result
        return None

    def first_not_none_of(self, transform: Callable[[T], Optional[R]]) -> R:
        """Alias for first_not_null_of() - more Pythonic naming."""
        return self.first_not_null_of(transform)

    def first_not_none_of_or_none(self, transform: Callable[[T], Optional[R]]) -> Optional[R]:
        """Alias for first_not_null_of_or_null() - more Pythonic naming."""
        return self.first_not_null_of_or_null(transform)

    # Aggregation methods
    def max_by(self, selector: Callable[[T], Any]) -> T:
        """Returns the first element yielding the largest value of the given function."""
        if self.is_empty():
            raise ValueError("Cannot find max of empty list")
        return max(self._elements, key=selector)

    def min_by(self, selector: Callable[[T], Any]) -> T:
        """Returns the first element yielding the smallest value of the given function."""
        if self.is_empty():
            raise ValueError("Cannot find min of empty list")
        return min(self._elements, key=selector)

    def max_of(self, selector: Callable[[T], Any]) -> Any:
        """Returns the largest value among all values produced by selector function."""
        if self.is_empty():
            raise ValueError("Cannot find max of empty list")
        return max(selector(element) for element in self._elements)

    def min_of(self, selector: Callable[[T], Any]) -> Any:
        """Returns the smallest value among all values produced by selector function."""
        if self.is_empty():
            raise ValueError("Cannot find min of empty list")
        return min(selector(element) for element in self._elements)

    def max_of_or_null(self, selector: Callable[[T], Any]) -> Optional[Any]:
        """Returns the largest value among all values produced by selector function or null if there are no elements."""
        if self.is_empty():
            return None
        return max(selector(element) for element in self._elements)

    def max_of_or_none(self, selector: Callable[[T], Any]) -> Optional[Any]:
        """Alias for max_of_or_null() - more Pythonic naming."""
        return self.max_of_or_null(selector)

    def min_of_or_null(self, selector: Callable[[T], Any]) -> Optional[Any]:
        """Returns the smallest value among all values produced by selector function or null if there are no elements."""
        if self.is_empty():
            return None
        return min(selector(element) for element in self._elements)

    def min_of_or_none(self, selector: Callable[[T], Any]) -> Optional[Any]:
        """Alias for min_of_or_null() - more Pythonic naming."""
        return self.min_of_or_null(selector)

    def max_of_with(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Any:
        """Returns the largest value according to the provided comparator among all values produced by selector function."""
        if self.is_empty():
            raise ValueError("Cannot find max of empty list")
        values = [selector(element) for element in self._elements]
        # Python doesn't have cmp parameter in max/min, so we need to use a different approach
        result = values[0]
        for value in values[1:]:
            if comparator(value, result) > 0:
                result = value
        return result

    def min_of_with(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Any:
        """Returns the smallest value according to the provided comparator among all values produced by selector function."""
        if self.is_empty():
            raise ValueError("Cannot find min of empty list")
        values = [selector(element) for element in self._elements]
        # Python doesn't have cmp parameter in max/min, so we need to use a different approach
        result = values[0]
        for value in values[1:]:
            if comparator(value, result) < 0:
                result = value
        return result

    def max_of_with_or_null(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Optional[Any]:
        """Returns the largest value according to the provided comparator among all values produced by selector function or null."""
        if self.is_empty():
            return None
        values = [selector(element) for element in self._elements]
        # Python doesn't have cmp parameter in max/min, so we need to use a different approach
        result = values[0]
        for value in values[1:]:
            if comparator(value, result) > 0:
                result = value
        return result

    def max_of_with_or_none(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Optional[Any]:
        """Alias for max_of_with_or_null() - more Pythonic naming."""
        return self.max_of_with_or_null(comparator, selector)

    def min_of_with_or_null(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Optional[Any]:
        """Returns the smallest value according to the provided comparator among all values produced by selector function or null."""
        if self.is_empty():
            return None
        values = [selector(element) for element in self._elements]
        # Python doesn't have cmp parameter in max/min, so we need to use a different approach
        result = values[0]
        for value in values[1:]:
            if comparator(value, result) < 0:
                result = value
        return result

    def min_of_with_or_none(self, comparator: Callable[[Any, Any], int], selector: Callable[[T], Any]) -> Optional[Any]:
        """Alias for min_of_with_or_null() - more Pythonic naming."""
        return self.min_of_with_or_null(comparator, selector)

    # Fold/Reduce variations
    def fold_indexed(self, initial: R, operation: Callable[[int, R, T], R]) -> R:
        """Accumulates value starting with initial value and applying operation from left to right to current accumulator value and each element with its index."""
        result = initial
        for i, element in enumerate(self._elements):
            result = operation(i, result, element)
        return result

    def fold_right(self, initial: R, operation: Callable[[T, R], R]) -> R:
        """Accumulates value starting with initial value and applying operation from right to left to each element and current accumulator value."""
        result = initial
        for element in reversed(self._elements):
            result = operation(element, result)
        return result

    def fold_right_indexed(self, initial: R, operation: Callable[[int, T, R], R]) -> R:
        """Accumulates value starting with initial value and applying operation from right to left to each element with its index and current accumulator value."""
        result = initial
        for i in range(len(self._elements) - 1, -1, -1):
            result = operation(i, self._elements[i], result)
        return result

    def reduce_indexed(self, operation: Callable[[int, T, T], T]) -> T:
        """Accumulates value starting with the first element and applying operation from left to right to current accumulator value and each element with its index."""
        if self.is_empty():
            raise ValueError("Cannot reduce empty list")
        result = self._elements[0]
        for i in range(1, len(self._elements)):
            result = operation(i, result, self._elements[i])
        return result

    def reduce_right(self, operation: Callable[[T, T], T]) -> T:
        """Accumulates value starting with the last element and applying operation from right to left to each element and current accumulator value."""
        if self.is_empty():
            raise ValueError("Cannot reduce empty list")
        result = self._elements[-1]
        for i in range(len(self._elements) - 2, -1, -1):
            result = operation(self._elements[i], result)
        return result

    def reduce_right_indexed(self, operation: Callable[[int, T, T], T]) -> T:
        """Accumulates value starting with the last element and applying operation from right to left to each element with its index and current accumulator value."""
        if self.is_empty():
            raise ValueError("Cannot reduce empty list")
        result = self._elements[-1]
        for i in range(len(self._elements) - 2, -1, -1):
            result = operation(i, self._elements[i], result)
        return result

    def reduce_indexed_or_null(self, operation: Callable[[int, T, T], T]) -> Optional[T]:
        """Accumulates value starting with the first element and applying operation from left to right, or null if the list is empty."""
        if self.is_empty():
            return None
        return self.reduce_indexed(operation)

    def reduce_indexed_or_none(self, operation: Callable[[int, T, T], T]) -> Optional[T]:
        """Alias for reduce_indexed_or_null() - more Pythonic naming."""
        return self.reduce_indexed_or_null(operation)

    def reduce_right_or_null(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Accumulates value starting with the last element and applying operation from right to left, or null if the list is empty."""
        if self.is_empty():
            return None
        return self.reduce_right(operation)

    def reduce_right_or_none(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Alias for reduce_right_or_null() - more Pythonic naming."""
        return self.reduce_right_or_null(operation)

    def reduce_right_indexed_or_null(self, operation: Callable[[int, T, T], T]) -> Optional[T]:
        """Accumulates value starting with the last element and applying operation from right to left with indices, or null if the list is empty."""
        if self.is_empty():
            return None
        return self.reduce_right_indexed(operation)

    def reduce_right_indexed_or_none(self, operation: Callable[[int, T, T], T]) -> Optional[T]:
        """Alias for reduce_right_indexed_or_null() - more Pythonic naming."""
        return self.reduce_right_indexed_or_null(operation)

    def reduce_or_null(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Accumulates value starting with the first element and applying operation from left to right, or null if the list is empty."""
        if self.is_empty():
            return None
        return self.reduce(operation)

    def reduce_or_none(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Alias for reduce_or_null() - more Pythonic naming."""
        return self.reduce_or_null(operation)

    def running_fold(self, initial: R, operation: Callable[[R, T], R]) -> 'KotList[R]':
        """Returns a list containing successive accumulation values generated by applying operation from left to right."""
        # This is the same as scan, which is already implemented
        return self.scan(initial, operation)

    def running_fold_indexed(self, initial: R, operation: Callable[[int, R, T], R]) -> 'KotList[R]':
        """Returns a list containing successive accumulation values generated by applying operation from left to right with indices."""
        result = [initial]
        acc = initial
        for i, element in enumerate(self._elements):
            acc = operation(i, acc, element)
            result.append(acc)
        return KotList(result)

    def scan_indexed(self, initial: R, operation: Callable[[int, R, T], R]) -> 'KotList[R]':
        """Returns a list containing successive accumulation values generated by applying operation from left to right with indices."""
        return self.running_fold_indexed(initial, operation)

    def running_reduce(self, operation: Callable[[T, T], T]) -> 'KotList[T]':
        """Returns a list containing successive accumulation values generated by applying operation from left to right."""
        if self.is_empty():
            return KotList()
        result = [self._elements[0]]
        acc = self._elements[0]
        for element in self._elements[1:]:
            acc = operation(acc, element)
            result.append(acc)
        return KotList(result)

    def running_reduce_indexed(self, operation: Callable[[int, T, T], T]) -> 'KotList[T]':
        """Returns a list containing successive accumulation values generated by applying operation from left to right with indices."""
        if self.is_empty():
            return KotList()
        result = [self._elements[0]]
        acc = self._elements[0]
        for i in range(1, len(self._elements)):
            acc = operation(i, acc, self._elements[i])
            result.append(acc)
        return KotList(result)

    # Other methods
    def as_reversed(self) -> 'KotList[T]':
        """Returns a reversed read-only view of the original List."""
        # In Python, we'll return a new KotList with reversed elements
        # since we don't have true "view" semantics like Kotlin
        return self.reversed()

    def as_sequence(self) -> Iterator[T]:
        """Creates a Sequence instance that wraps the original list."""
        # In Python, we'll return an iterator
        return iter(self._elements)

    def with_index(self) -> Iterator[Tuple[int, T]]:
        """Returns a lazy Iterable of IndexedValue for each element of the original list."""
        return enumerate(self._elements)

    def on_each_indexed(self, action: Callable[[int, T], None]) -> 'KotList[T]':
        """Performs the given action on each element with its index, returning the list itself afterwards."""
        for i, element in enumerate(self._elements):
            action(i, element)
        return self

    def if_empty(self, default_value: Callable[[], 'KotList[T]']) -> 'KotList[T]':
        """Returns this list if it's not empty or the result of calling defaultValue function if the list is empty."""
        return self if self.is_not_empty() else default_value()

    # ListIterator methods
    def list_iterator(self, index: int = 0) -> Iterator[T]:
        """Returns a list iterator over the elements in this list."""
        if index < 0 or index > self.size:
            raise IndexError(f"Index {index} out of bounds for list of size {self.size}")
        return iter(self._elements[index:])

    # Operator overloads
    def __add__(self, other: Union[T, Iterable[T]]) -> 'KotList[T]':
        """Overload + operator for plus() method."""
        return self.plus(other)

    def __sub__(self, other: Union[T, Iterable[T]]) -> 'KotList[T]':
        """Overload - operator for minus() method."""
        return self.minus(other)



"""
KotMutableMap: A Python implementation of Kotlin's MutableMap interface with snake_case naming convention.
"""

from __future__ import annotations

from typing import TypeVar, Dict, List, Iterator, Optional, Callable, Tuple, Type

from kotcollections.kot_map import KotMap

K = TypeVar('K')
V = TypeVar('V')



class KotMutableMap(KotMap[K, V]):
    """A Python implementation of Kotlin's MutableMap interface.
    
    This class extends KotMap with mutation methods, providing full
    Kotlin MutableMap functionality with snake_case naming.
    """

    def __init__(self, elements: Optional[Dict[K, V] | List[Tuple[K, V]] | Iterator[Tuple[K, V]]] = None):
        """Initialize a KotMutableMap with optional elements.
        
        Args:
            elements: Initial elements for the map (dict, list of tuples, or iterator of tuples)
        """
        super().__init__(elements)

    @classmethod
    def __class_getitem__(cls, types: Tuple[Type[K], Type[V]]) -> Type['KotMutableMap[K, V]']:
        """Enable KotMutableMap[KeyType, ValueType]() syntax for type specification.
        
        Example:
            animals_by_name = KotMutableMap[str, Animal]()
            animals_by_name.put("Buddy", Dog("Buddy"))
            animals_by_name.put("Whiskers", Cat("Whiskers"))
        """
        key_type, value_type = types
        
        class TypedKotMutableMap(cls):
            def __init__(self, elements=None):
                # Only set types if they are actual types, not type variables
                self._elements = {}
                self._key_type = key_type if isinstance(key_type, type) else None
                self._value_type = value_type if isinstance(value_type, type) else None
                # Now process elements with the correct types set
                if elements is not None:
                    if isinstance(elements, dict):
                        for key, value in elements.items():
                            self._put_with_type_check(key, value)
                    elif isinstance(elements, list):
                        for key, value in elements:
                            self._put_with_type_check(key, value)
                    else:  # Iterator
                        for key, value in elements:
                            self._put_with_type_check(key, value)
        
        # Set a meaningful name for debugging
        TypedKotMutableMap.__name__ = f"{cls.__name__}[{key_type.__name__}, {value_type.__name__}]"
        TypedKotMutableMap.__qualname__ = f"{cls.__qualname__}[{key_type.__name__}, {value_type.__name__}]"
        
        return TypedKotMutableMap

    @classmethod
    def of_type(cls, key_type: Type[K], value_type: Type[V], elements: Optional[Dict[K, V] | List[Tuple[K, V]] | Iterator[Tuple[K, V]]] = None) -> 'KotMutableMap[K, V]':
        """Create a KotMutableMap with specific key and value types.
        
        This is useful when you want to create a mutable map with parent types
        but only have instances of child types.
        
        Args:
            key_type: The type of keys this map will contain
            value_type: The type of values this map will contain
            elements: Optional initial elements
            
        Returns:
            A new KotMutableMap instance with the specified types
            
        Example:
            animals_by_name = KotMutableMap.of_type(str, Animal, [("Buddy", Dog("Buddy")), ("Whiskers", Cat("Whiskers"))])
            # or empty map
            animals_by_name = KotMutableMap.of_type(str, Animal)
            animals_by_name.put("Max", Dog("Max"))
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[key_type, value_type]
        return typed_class(elements)

    # Mutation operations

    def put(self, key: K, value: V) -> Optional[V]:
        """Associates the specified value with the specified key in the map.
        
        Returns:
            The previous value associated with the key, or null if the key was not present.
        """
        old_value = self._elements.get(key)
        self._put_with_type_check(key, value)
        return old_value

    def put_all(self, from_map: Dict[K, V] | 'KotMap[K, V]' | List[Tuple[K, V]]) -> None:
        """Updates this map with key/value pairs from the specified map."""
        if isinstance(from_map, KotMap):
            from_map = from_map._elements
        elif isinstance(from_map, list):
            from_map = dict(from_map)
        
        for key, value in from_map.items():
            self._put_with_type_check(key, value)

    def put_if_absent(self, key: K, value: V) -> Optional[V]:
        """Associates the specified value with the specified key only if it is not already associated.
        
        Returns:
            The current value associated with the key, or null if none and the new value was added.
        """
        if key in self._elements:
            return self._elements[key]
        self._put_with_type_check(key, value)
        return None

    def remove(self, key: K) -> Optional[V]:
        """Removes the specified key and its corresponding value from this map.
        
        Returns:
            The value associated with the key, or null if the key was not present.
        """
        if key in self._elements:
            value = self._elements.pop(key)
            if self.is_empty():
                self._key_type = None
                self._value_type = None
            return value
        return None

    def remove_value(self, key: K, value: V) -> bool:
        """Removes the entry for the specified key only if it is mapped to the specified value.
        
        Returns:
            true if the entry was removed.
        """
        if self.get(key) == value:
            self.remove(key)
            return True
        return False

    def clear(self) -> None:
        """Removes all key/value pairs from the map."""
        self._elements.clear()
        self._key_type = None
        self._value_type = None

    # Advanced mutation operations

    def get_or_put(self, key: K, default_value: Callable[[], V]) -> V:
        """Returns the value for the given key. If the key is not found, calls defaultValue and puts its result."""
        if key in self._elements:
            return self._elements[key]
        value = default_value()
        self._put_with_type_check(key, value)
        return value

    def compute(self, key: K, remapping_function: Callable[[K, Optional[V]], Optional[V]]) -> Optional[V]:
        """Attempts to compute a mapping for the specified key and its current mapped value.
        
        The remapping function receives the key and current value (or null if not present).
        If it returns null, the mapping is removed (if it was present).
        """
        current_value = self._elements.get(key)
        new_value = remapping_function(key, current_value)
        
        if new_value is None:
            if key in self._elements:
                self.remove(key)
        else:
            self._put_with_type_check(key, new_value)
        
        return new_value

    def compute_if_absent(self, key: K, mapping_function: Callable[[K], V]) -> V:
        """If the specified key is not already associated with a value, computes its value using the given function.
        
        Returns:
            The current (existing or computed) value associated with the specified key.
        """
        if key in self._elements:
            return self._elements[key]
        
        value = mapping_function(key)
        if value is not None:
            self._put_with_type_check(key, value)
        return value

    def compute_if_present(self, key: K, remapping_function: Callable[[K, V], Optional[V]]) -> Optional[V]:
        """If the value for the specified key is present, computes a new mapping.
        
        The remapping function receives the key and current value.
        If it returns null, the mapping is removed.
        """
        if key not in self._elements:
            return None
        
        current_value = self._elements[key]
        new_value = remapping_function(key, current_value)
        
        if new_value is None:
            self.remove(key)
        else:
            self._put_with_type_check(key, new_value)
        
        return new_value

    def replace(self, key: K, value: V) -> Optional[V]:
        """Replaces the entry for the specified key only if it is currently mapped to some value.
        
        Returns:
            The previous value associated with the key, or null if the key was not present.
        """
        if key in self._elements:
            return self.put(key, value)
        return None

    def replace_all(self, transform: Callable[[K, V], V]) -> None:
        """Replaces each entry's value with the result of invoking the given function on that entry."""
        for key in list(self._elements.keys()):
            current_value = self._elements[key]
            new_value = transform(key, current_value)
            self._put_with_type_check(key, new_value)

    def merge(self, key: K, value: V, remapping_function: Callable[[V, V], V]) -> V:
        """If the specified key is not already associated with a value or is associated with null, associates it with the given value.
        Otherwise, replaces the value with the results of the given remapping function.
        
        Returns:
            The new value associated with the specified key.
        """
        if key not in self._elements:
            self._put_with_type_check(key, value)
            return value
        
        old_value = self._elements[key]
        new_value = remapping_function(old_value, value)
        self._put_with_type_check(key, new_value)
        return new_value

    # Bulk operations from other maps

    def plus_assign(self, other: Dict[K, V] | 'KotMap[K, V]' | Tuple[K, V]) -> None:
        """Adds all entries from the specified map or the specified pair to this map."""
        if isinstance(other, tuple):
            key, value = other
            self.put(key, value)
        else:
            self.put_all(other)

    def minus_assign(self, key: K) -> None:
        """Removes the specified key from this map."""
        self.remove(key)

    # Python special methods for mutable operations

    def __setitem__(self, key: K, value: V) -> None:
        """Set value by key using [] operator."""
        self.put(key, value)

    def __delitem__(self, key: K) -> None:
        """Delete entry by key using del operator."""
        if key not in self._elements:
            raise KeyError(f"Key {key} not found in map")
        self.remove(key)

    def pop(self, key: K, default: Optional[V] = None) -> Optional[V]:
        """Remove specified key and return the corresponding value.
        If key is not found, default is returned if given, otherwise KeyError is raised."""
        if key in self._elements:
            return self.remove(key)
        if default is not None:
            return default
        raise KeyError(f"Key {key} not found in map")

    def popitem(self) -> Tuple[K, V]:
        """Remove and return an arbitrary (key, value) pair from the map."""
        if self.is_empty():
            raise KeyError("Map is empty")
        key, value = next(iter(self._elements.items()))
        self.remove(key)
        return key, value

    def update(self, other: Dict[K, V] | 'KotMap[K, V]' | List[Tuple[K, V]]) -> None:
        """Update the map with key/value pairs from other, overwriting existing keys."""
        self.put_all(other)

    # Override __hash__ to make it unhashable (mutable objects shouldn't be hashable)
    __hash__ = None  # type: ignore


"""
KotMap: A Python implementation of Kotlin's Map interface with snake_case naming convention.
"""

from __future__ import annotations

from typing import TypeVar, Generic, Callable, Optional, Dict, Iterator, Any, Tuple, List, Set, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from kotcollections import KotMutableMap, KotList, KotSet

K = TypeVar('K')
V = TypeVar('V')
R = TypeVar('R')



class KotMap(Generic[K, V]):
    """A Python implementation of Kotlin's Map interface.
    
    This class provides all methods from Kotlin's Map interface with snake_case naming,
    maintaining type safety and immutability.
    """

    def __init__(self, elements: Optional[Dict[K, V] | List[Tuple[K, V]] | Iterator[Tuple[K, V]]] = None):
        """Initialize a KotMap with optional elements.
        
        Args:
            elements: Initial elements for the map (dict, list of tuples, or iterator of tuples)
        """
        self._elements: Dict[K, V] = {}
        self._key_type: Optional[type] = None
        self._value_type: Optional[type] = None

        if elements is not None:
            if isinstance(elements, dict):
                for key, value in elements.items():
                    self._put_with_type_check(key, value)
            elif isinstance(elements, list):
                for key, value in elements:
                    self._put_with_type_check(key, value)
            else:  # Iterator
                for key, value in elements:
                    self._put_with_type_check(key, value)

    @classmethod
    def __class_getitem__(cls, types: Tuple[Type[K], Type[V]]) -> Type['KotMap[K, V]']:
        """Enable KotMap[KeyType, ValueType]() syntax for type specification.
        
        Example:
            animals_by_name = KotMap[str, Animal]()
            animals_by_name.put("Buddy", Dog("Buddy"))
            animals_by_name.put("Whiskers", Cat("Whiskers"))
        """
        key_type, value_type = types
        
        class TypedKotMap(cls):
            def __init__(self, elements=None):
                # Only set types if they are actual types, not type variables
                self._elements = {}
                self._key_type = key_type if isinstance(key_type, type) else None
                self._value_type = value_type if isinstance(value_type, type) else None
                # Now process elements with the correct types set
                if elements is not None:
                    if isinstance(elements, dict):
                        for key, value in elements.items():
                            self._put_with_type_check(key, value)
                    elif isinstance(elements, list):
                        for key, value in elements:
                            self._put_with_type_check(key, value)
                    else:  # Iterator
                        for key, value in elements:
                            self._put_with_type_check(key, value)
        
        # Set a meaningful name for debugging
        TypedKotMap.__name__ = f"{cls.__name__}[{key_type.__name__}, {value_type.__name__}]"
        TypedKotMap.__qualname__ = f"{cls.__qualname__}[{key_type.__name__}, {value_type.__name__}]"
        
        return TypedKotMap

    @classmethod
    def of_type(cls, key_type: Type[K], value_type: Type[V], elements: Optional[Dict[K, V] | List[Tuple[K, V]] | Iterator[Tuple[K, V]]] = None) -> 'KotMap[K, V]':
        """Create a KotMap with specific key and value types.
        
        This is useful when you want to create a map with parent types
        but only have instances of child types.
        
        Args:
            key_type: The type of keys this map will contain
            value_type: The type of values this map will contain
            elements: Optional initial elements
            
        Returns:
            A new KotMap instance with the specified types
            
        Example:
            animals_by_name = KotMap.of_type(str, Animal, [("Buddy", Dog("Buddy")), ("Whiskers", Cat("Whiskers"))])
            # or empty map
            animals_by_name = KotMap.of_type(str, Animal)
            animals_by_name.put("Max", Dog("Max"))
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[key_type, value_type]
        return typed_class(elements)

    def _put_with_type_check(self, key: K, value: V) -> None:
        """Add a key-value pair with type checking."""
        # Skip type checking if key_type or value_type is a type variable or not a real type
        if (self._key_type is not None and not isinstance(self._key_type, type)) or \
           (self._value_type is not None and not isinstance(self._value_type, type)):
            self._elements[key] = value
            return
            
        # Set key type
        if self._key_type is None and key is not None:
            self._key_type = type(key) if not isinstance(key, KotMap) else KotMap

        # Set value type
        if self._value_type is None and value is not None:
            self._value_type = type(value) if not isinstance(value, KotMap) else KotMap

        # Check key type
        if key is not None and self._key_type is not None:
            if isinstance(key, KotMap) and self._key_type == KotMap:
                pass  # KotMap keys are allowed
            elif not isinstance(key, KotMap):
                # Check if key is an instance of the expected type
                if isinstance(key, self._key_type):
                    pass  # Direct instance check passed
                # Special handling for __class_getitem__ types (e.g., KotList[Holiday])
                elif (hasattr(self._key_type, '__base__') and 
                      hasattr(self._key_type, '__name__') and 
                      '[' in self._key_type.__name__ and
                      isinstance(key, self._key_type.__base__)):
                    # Check if the key has matching element type for KotList/KotSet/KotMap types
                    if hasattr(key, '_element_type') and hasattr(self._key_type, '__new__'):
                        # Extract expected element type from the __class_getitem__ type
                        # This is a more strict check for collection types
                        pass
                    else:
                        pass
                else:
                    raise TypeError(f"All keys must be of type {self._key_type.__name__}, got {type(key).__name__}")

        # Check value type
        if value is not None and self._value_type is not None:
            if isinstance(value, KotMap) and self._value_type == KotMap:
                pass  # KotMap values are allowed
            elif not isinstance(value, KotMap):
                # Check if value is an instance of the expected type
                if isinstance(value, self._value_type):
                    pass  # Direct instance check passed
                # Special handling for __class_getitem__ types (e.g., KotList[Holiday])
                elif (hasattr(self._value_type, '__base__') and 
                      hasattr(self._value_type, '__name__') and 
                      '[' in self._value_type.__name__ and
                      isinstance(value, self._value_type.__base__)):
                    # Check if the value has matching element type for KotList/KotSet/KotMap types
                    if hasattr(value, '_element_type') and hasattr(self._value_type, '__new__'):
                        # Extract expected element type from the __class_getitem__ type
                        # This is a more strict check for collection types
                        pass
                    else:
                        pass
                else:
                    raise TypeError(f"All values must be of type {self._value_type.__name__}, got {type(value).__name__}")

        self._elements[key] = value

    # Basic Map operations

    def is_empty(self) -> bool:
        """Returns true if the map is empty."""
        return len(self._elements) == 0

    def is_not_empty(self) -> bool:
        """Returns true if the map is not empty."""
        return len(self._elements) > 0

    @property
    def size(self) -> int:
        """Returns the size of the map."""
        return len(self._elements)

    def contains_key(self, key: K) -> bool:
        """Returns true if the map contains the specified key."""
        return key in self._elements

    def contains_value(self, value: V) -> bool:
        """Returns true if the map maps one or more keys to the specified value."""
        return value in self._elements.values()

    # Access operations

    def get(self, key: K) -> Optional[V]:
        """Returns the value corresponding to the given key, or null if such a key is not present in the map."""
        return self._elements.get(key)

    def get_or_default(self, key: K, default_value: V) -> V:
        """Returns the value corresponding to the given key, or default_value if such a key is not present in the map."""
        return self._elements.get(key, default_value)

    def get_or_else(self, key: K, default_value: Callable[[], V]) -> V:
        """Returns the value for the given key. If the key is not found, calls the defaultValue function."""
        if key in self._elements:
            return self._elements[key]
        return default_value()

    def get_or_null(self, key: K) -> Optional[V]:
        """Returns the value corresponding to the given key, or null if such a key is not present in the map."""
        return self._elements.get(key)

    def get_or_none(self, key: K) -> Optional[V]:
        """Pythonic alias for get_or_null()."""
        return self.get_or_null(key)

    def get_value(self, key: K) -> V:
        """Returns the value for the given key or throws an exception if the key is missing in the map."""
        if key not in self._elements:
            raise KeyError(f"Key {key} is missing in the map.")
        return self._elements[key]

    # Collection views

    @property
    def keys(self) -> 'KotSet[K]':
        """Returns a read-only Set of all keys in this map."""
        from kotcollections import KotSet
        return KotSet(list(self._elements.keys()))

    @property
    def values(self) -> 'KotList[V]':
        """Returns a read-only Collection of all values in this map."""
        from kotcollections import KotList
        return KotList(self._elements.values())

    @property
    def entries(self) -> 'KotSet[Tuple[K, V]]':
        """Returns a read-only Set of all key/value pairs in this map."""
        from kotcollections import KotSet
        return KotSet(list(self._elements.items()))

    # Checking operations

    def all(self, predicate: Callable[[K, V], bool]) -> bool:
        """Returns true if all entries match the given predicate."""
        return all(predicate(k, v) for k, v in self._elements.items())

    def any(self, predicate: Callable[[K, V], bool]) -> bool:
        """Returns true if at least one entry matches the given predicate."""
        return any(predicate(k, v) for k, v in self._elements.items())

    def none(self, predicate: Callable[[K, V], bool]) -> bool:
        """Returns true if no entries match the given predicate."""
        return not any(predicate(k, v) for k, v in self._elements.items())

    def count(self, predicate: Optional[Callable[[K, V], bool]] = None) -> int:
        """Returns the number of entries matching the given predicate."""
        if predicate is None:
            return self.size
        return sum(1 for k, v in self._elements.items() if predicate(k, v))

    # Filtering operations

    def filter(self, predicate: Callable[[K, V], bool]) -> 'KotMap[K, V]':
        """Returns a map containing only entries matching the given predicate."""
        filtered_pairs = [(k, v) for k, v in self._elements.items() if predicate(k, v)]
        return KotMap(filtered_pairs)

    def filter_not(self, predicate: Callable[[K, V], bool]) -> 'KotMap[K, V]':
        """Returns a map containing only entries not matching the given predicate."""
        filtered_pairs = [(k, v) for k, v in self._elements.items() if not predicate(k, v)]
        return KotMap(filtered_pairs)

    def filter_keys(self, predicate: Callable[[K], bool]) -> 'KotMap[K, V]':
        """Returns a map containing all entries with keys matching the given predicate."""
        filtered_pairs = [(k, v) for k, v in self._elements.items() if predicate(k)]
        return KotMap(filtered_pairs)

    def filter_values(self, predicate: Callable[[V], bool]) -> 'KotMap[K, V]':
        """Returns a map containing all entries with values matching the given predicate."""
        filtered_pairs = [(k, v) for k, v in self._elements.items() if predicate(v)]
        return KotMap(filtered_pairs)

    def filter_not_null(self) -> 'KotMap[K, V]':
        """Returns a map containing only entries with non-null values."""
        filtered_pairs = [(k, v) for k, v in self._elements.items() if v is not None]
        return KotMap(filtered_pairs)

    def filter_not_none(self) -> 'KotMap[K, V]':
        """Pythonic alias for filter_not_null()."""
        return self.filter_not_null()

    # Transformation operations

    def map(self, transform: Callable[[K, V], R]) -> 'KotList[R]':
        """Returns a list containing the results of applying the given transform function to each entry."""
        from kotcollections import KotList
        return KotList([transform(k, v) for k, v in self._elements.items()])

    def map_keys(self, transform: Callable[[K, V], R]) -> 'KotMap[R, V]':
        """Returns a new map with entries having the keys obtained by applying the transform function to each entry."""
        transformed_pairs = [(transform(k, v), v) for k, v in self._elements.items()]
        return KotMap(transformed_pairs)

    def map_values(self, transform: Callable[[K, V], R]) -> 'KotMap[K, R]':
        """Returns a new map with entries having the values obtained by applying the transform function to each entry."""
        transformed_pairs = [(k, transform(k, v)) for k, v in self._elements.items()]
        return KotMap(transformed_pairs)

    def map_not_null(self, transform: Callable[[K, V], Optional[R]]) -> 'KotList[R]':
        """Returns a list containing only the non-null results of applying the given transform function."""
        from kotcollections import KotList
        results = []
        for k, v in self._elements.items():
            result = transform(k, v)
            if result is not None:
                results.append(result)
        return KotList(results)

    def map_not_none(self, transform: Callable[[K, V], Optional[R]]) -> 'KotList[R]':
        """Pythonic alias for map_not_null()."""
        return self.map_not_null(transform)

    def flat_map(self, transform: Callable[[K, V], Iterator[R]]) -> 'KotList[R]':
        """Returns a single list of all elements yielded from results of transform function."""
        from kotcollections import KotList
        results = []
        for k, v in self._elements.items():
            results.extend(transform(k, v))
        return KotList(results)

    # Conversion operations

    def to_list(self) -> List[Tuple[K, V]]:
        """Returns a List containing all key-value pairs."""
        return list(self._elements.copy().items())

    def to_dict(self) -> Dict[K, V]:
        """Returns a Python dict containing all key-value pairs."""
        return dict(self._elements.copy())

    def to_kot_map(self) -> 'KotMap[K, V]':
        """Returns a KotMap containing all key-value pairs."""
        # Preserve type information when converting
        if self._key_type is not None and self._value_type is not None:
            return KotMap.of_type(self._key_type, self._value_type, self._elements.copy())
        else:
            return KotMap(self._elements.copy())

    def to_kot_mutable_map(self) -> 'KotMutableMap[K, V]':
        """Returns a KotMutableMap containing all key-value pairs."""
        from kotcollections.kot_mutable_map import KotMutableMap
        # Preserve type information when converting
        if self._key_type is not None and self._value_type is not None:
            mutable_map = KotMutableMap.of_type(self._key_type, self._value_type, self._elements.copy())
        else:
            mutable_map = KotMutableMap(self._elements.copy())
        return mutable_map

    # Action operations

    def for_each(self, action: Callable[[K, V], None]) -> None:
        """Performs the given action on each entry."""
        for k, v in self._elements.items():
            action(k, v)

    def on_each(self, action: Callable[[K, V], None]) -> 'KotMap[K, V]':
        """Performs the given action on each entry and returns the map itself."""
        for k, v in self._elements.items():
            action(k, v)
        return self

    # Finding operations

    def max_by(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Returns the entry yielding the largest value of the given function."""
        if self.is_empty():
            return None
        return max(self._elements.items(), key=lambda item: selector(item[0], item[1]))

    def max_by_or_null(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Returns the entry yielding the largest value of the given function or null if there are no entries."""
        return self.max_by(selector)

    def max_by_or_none(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Pythonic alias for max_by_or_null()."""
        return self.max_by_or_null(selector)

    def min_by(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Returns the entry yielding the smallest value of the given function."""
        if self.is_empty():
            return None
        return min(self._elements.items(), key=lambda item: selector(item[0], item[1]))

    def min_by_or_null(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Returns the entry yielding the smallest value of the given function or null if there are no entries."""
        return self.min_by(selector)

    def min_by_or_none(self, selector: Callable[[K, V], Any]) -> Optional[Tuple[K, V]]:
        """Pythonic alias for min_by_or_null()."""
        return self.min_by_or_null(selector)

    # String representation

    def join_to_string(
        self, separator: str = ", ", prefix: str = "", postfix: str = "",
        limit: int = -1, truncated: str = "...",
        transform: Optional[Callable[[K, V], str]] = None
    ) -> str:
        """Creates a string from all the entries separated using separator."""
        if transform is None:
            transform = lambda k, v: f"{k}={v}"

        items = []
        for i, (k, v) in enumerate(self._elements.items()):
            if limit >= 0 and i >= limit:
                items.append(truncated)
                break
            items.append(transform(k, v))

        return prefix + separator.join(items) + postfix

    # Operator methods

    def plus(self, other: 'KotMap[K, V]' | Dict[K, V] | Tuple[K, V]) -> 'KotMap[K, V]':
        """Returns a map containing all entries from this map and the given map/pair.
        
        If a key is present in both maps, the value from the other map will be used.
        """
        from kotcollections import KotList
        
        new_elements = dict(self._elements)

        if isinstance(other, tuple):
            key, value = other
            new_elements[key] = value
        elif isinstance(other, KotMap):
            new_elements.update(other._elements)
        elif isinstance(other, KotList):
            # Support KotList of pairs
            for item in other:
                if isinstance(item, tuple) and len(item) == 2:
                    key, value = item
                    new_elements[key] = value
                else:
                    raise ValueError(f"Expected a tuple of (key, value), got {type(item)}")
        else:  # Dict
            new_elements.update(other)

        return KotMap(new_elements)

    def minus(self, key: K | List[K] | Set[K]) -> 'KotMap[K, V]':
        """Returns a map containing all entries from this map except those whose keys are in the given collection."""
        from kotcollections import KotSet
        
        if isinstance(key, (list, set)):
            new_elements = {k: v for k, v in self._elements.items() if k not in key}
        elif isinstance(key, KotSet):
            # Convert KotSet to set for efficient lookup
            key_set = set(key)
            new_elements = {k: v for k, v in self._elements.items() if k not in key_set}
        else:
            new_elements = {k: v for k, v in self._elements.items() if k != key}

        return KotMap(new_elements)

    def with_default(self, default_value: Callable[[K], V]) -> 'KotMapWithDefault[K, V]':
        """Returns a wrapper of this map having the implicit default value provided by the specified function."""
        return KotMapWithDefault(self, default_value)

    # Python special methods

    def __len__(self) -> int:
        """Return the number of entries in the map."""
        return self.size

    def __contains__(self, key: K) -> bool:
        """Check if a key is in the map."""
        return self.contains_key(key)

    def __getitem__(self, key: K) -> V:
        """Get value by key using [] operator."""
        if key not in self._elements:
            raise KeyError(f"Key {key} not found in map")
        return self._elements[key]

    def __iter__(self) -> Iterator[K]:
        """Iterate over keys."""
        return iter(self._elements)

    def __repr__(self) -> str:
        """Return string representation of the map."""
        return f"KotMap({dict(self._elements)})"

    def __eq__(self, other: Any) -> bool:
        """Check equality with another KotMap."""
        if not isinstance(other, KotMap):
            return False
        return self._elements == other._elements

    def __add__(self, other: 'KotMap[K, V]' | Dict[K, V] | Tuple[K, V]) -> 'KotMap[K, V]':
        """Implement + operator for maps."""
        return self.plus(other)

    def __sub__(self, key: K | List[K] | Set[K]) -> 'KotMap[K, V]':
        """Implement - operator for maps."""
        return self.minus(key)

    def __hash__(self) -> int:
        """Return hash of the map."""
        # Since KotMap is immutable, we can cache the hash
        if not hasattr(self, '_cached_hash'):
            items = tuple(sorted(self._elements.items()))
            self._cached_hash = hash(items)
        return self._cached_hash


class KotMapWithDefault(KotMap[K, V]):
    """A wrapper class that provides a default value for missing keys.
    
    This class wraps a KotMap and provides a default value function that
    is called when a key is not found in the map.
    """

    def __init__(self, map: KotMap[K, V], default_value: Callable[[K], V]):
        """Initialize with a map and a default value function.
        
        Args:
            map: The underlying KotMap
            default_value: Function that provides default values for missing keys
        """
        # Copy the elements from the original map
        super().__init__(map._elements)
        self._default_value_function = default_value

    def get(self, key: K) -> V:
        """Returns the value corresponding to the given key, or the default value if the key is not present."""
        if key in self._elements:
            return self._elements[key]
        return self._default_value_function(key)

    def get_or_default(self, key: K, default_value: V) -> V:
        """Returns the value corresponding to the given key, or default_value if such a key is not present."""
        # This overrides the parent method to use the map's default if no explicit default is needed
        return self.get(key)

    def get_or_else(self, key: K, default_value: Callable[[], V]) -> V:
        """Returns the value for the given key. If the key is not found, calls the defaultValue function."""
        if key in self._elements:
            return self._elements[key]
        # Use the map's default value function, not the provided one
        return self._default_value_function(key)

    def get_value(self, key: K) -> V:
        """Returns the value for the given key, using the default if the key is missing."""
        return self.get(key)

    def __getitem__(self, key: K) -> V:
        """Get value by key using [] operator, with default value support."""
        return self.get(key)

    def __repr__(self) -> str:
        """Return string representation of the map with default."""
        return f"KotMapWithDefault({dict(self._elements)}, default_function)"



"""
KotMutableSet: A Python implementation of Kotlin's MutableSet interface with snake_case naming convention.
"""

from __future__ import annotations

from typing import TypeVar, Set, List, Iterator, Optional, Callable, Type, TYPE_CHECKING, Dict, Tuple

from kotcollections.kot_set import KotSet

if TYPE_CHECKING:
    from kotcollections.kot_list import KotList
    from kotcollections.kot_mutable_list import KotMutableList

T = TypeVar('T')



class KotMutableSet(KotSet[T]):
    """A Python implementation of Kotlin's MutableSet interface.
    
    This class extends KotSet with mutation methods, providing full
    Kotlin MutableSet functionality with snake_case naming.
    """

    def __init__(self, elements: Optional[Set[T] | List[T] | Iterator[T]] = None):
        """Initialize a KotMutableSet with optional elements.
        
        Args:
            elements: Initial elements for the set (set, list, or iterator)
        """
        super().__init__(elements)

    @classmethod
    def __class_getitem__(cls, element_type: Type[T]) -> Type['KotMutableSet[T]']:
        """Enable KotMutableSet[Type]() syntax for type specification.
        
        Example:
            animals = KotMutableSet[Animal]()
            animals.add(Dog("Buddy"))
            animals.add(Cat("Whiskers"))
        """
        class TypedKotMutableSet(cls):
            def __init__(self, elements=None):
                # Only set element type if it's an actual type, not a type variable
                if isinstance(element_type, type):
                    self._element_type = element_type
                else:
                    self._element_type = None
                self._elements = set()
                # Now process elements with the correct type set
                if elements is not None:
                    if hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
                        # It's a KotList or KotMutableList
                        for element in elements:
                            self._add_with_type_check(element)
                    elif isinstance(elements, set):
                        for element in elements:
                            self._add_with_type_check(element)
                    elif isinstance(elements, list):
                        for element in elements:
                            self._add_with_type_check(element)
                    else:  # Iterator
                        for element in elements:
                            self._add_with_type_check(element)
        
        # Set a meaningful name for debugging
        TypedKotMutableSet.__name__ = f"{cls.__name__}[{element_type.__name__}]"
        TypedKotMutableSet.__qualname__ = f"{cls.__qualname__}[{element_type.__name__}]"
        
        return TypedKotMutableSet

    @classmethod
    def of_type(cls, element_type: Type[T], elements: Optional[Set[T] | List[T] | Iterator[T]] = None) -> 'KotMutableSet[T]':
        """Create a KotMutableSet with a specific element type.
        
        This is useful when you want to create a mutable set of a parent type
        but only have instances of child types.
        
        Args:
            element_type: The type of elements this set will contain
            elements: Optional initial elements
            
        Returns:
            A new KotMutableSet instance with the specified element type
            
        Example:
            animals = KotMutableSet.of_type(Animal, [Dog("Buddy"), Cat("Whiskers")])
            # or empty set
            animals = KotMutableSet.of_type(Animal)
            animals.add(Dog("Max"))
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[element_type]
        return typed_class(elements)

    # Mutation operations

    def add(self, element: T) -> bool:
        """Adds the specified element to the set.
        
        Returns:
            true if the element has been added, false if the element is already in this set.
        """
        if element in self._elements:
            return False
        self._add_with_type_check(element)
        return True

    def add_all(self, elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> bool:
        """Adds all of the elements in the specified collection to this set.
        
        Returns:
            true if any of the specified elements was added to the set.
        """
        if isinstance(elements, KotSet):
            elements = elements._elements
        elif hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
            # It's a KotList or KotMutableList
            elements = set(elements)
        elif isinstance(elements, list):
            elements = set(elements)

        initial_size = self.size
        for element in elements:
            if element not in self._elements:
                self._add_with_type_check(element)
        return self.size > initial_size

    def remove(self, element: T) -> bool:
        """Removes a single instance of the specified element from this set.
        
        Returns:
            true if the element has been successfully removed.
        """
        if element in self._elements:
            self._elements.remove(element)
            if self.is_empty():
                self._element_type = None
            return True
        return False

    def remove_all(self, elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> bool:
        """Removes all of this set's elements that are also contained in the specified collection.
        
        Returns:
            true if any of the specified elements was removed from the set.
        """
        if isinstance(elements, KotSet):
            elements = elements._elements
        elif hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
            # It's a KotList or KotMutableList
            elements = set(elements)
        elif isinstance(elements, list):
            elements = set(elements)

        initial_size = self.size
        self._elements.difference_update(elements)
        if self.is_empty():
            self._element_type = None
        return self.size < initial_size

    def retain_all(self, elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> bool:
        """Retains only the elements in this set that are contained in the specified collection.
        
        Returns:
            true if any element was removed from the set.
        """
        if isinstance(elements, KotSet):
            elements = elements._elements
        elif hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
            # It's a KotList or KotMutableList
            elements = set(elements)
        elif isinstance(elements, list):
            elements = set(elements)

        initial_size = self.size
        self._elements.intersection_update(elements)
        if self.is_empty():
            self._element_type = None
        return self.size < initial_size

    def clear(self) -> None:
        """Removes all elements from this set."""
        self._elements.clear()
        self._element_type = None

    # Additional mutation operations

    def remove_if(self, predicate: Callable[[T], bool]) -> bool:
        """Removes all elements that match the given predicate.
        
        Returns:
            true if any elements were removed.
        """
        to_remove = [element for element in self._elements if predicate(element)]
        if to_remove:
            for element in to_remove:
                self._elements.remove(element)
            if self.is_empty():
                self._element_type = None
            return True
        return False

    def retain_if(self, predicate: Callable[[T], bool]) -> bool:
        """Retains only elements that match the given predicate.
        
        Returns:
            true if any elements were removed.
        """
        to_remove = [element for element in self._elements if not predicate(element)]
        if to_remove:
            for element in to_remove:
                self._elements.remove(element)
            if self.is_empty():
                self._element_type = None
            return True
        return False

    # Union/Intersection/Difference with mutation

    def union_update(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> None:
        """Adds all elements from the other collection to this set."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        for element in other:
            if element not in self._elements:
                self._add_with_type_check(element)

    def intersect_update(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> None:
        """Retains only elements that are contained in the specified collection."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        self._elements.intersection_update(other)
        if self.is_empty():
            self._element_type = None

    def subtract_update(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> None:
        """Removes all elements that are contained in the specified collection."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        self._elements.difference_update(other)
        if self.is_empty():
            self._element_type = None

    # Operator overloads for mutation

    def __iadd__(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotMutableSet[T]':
        """In-place union using += operator."""
        self.union_update(other)
        return self

    def __isub__(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotMutableSet[T]':
        """In-place subtraction using -= operator."""
        self.subtract_update(other)
        return self

    def __iand__(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotMutableSet[T]':
        """In-place intersection using &= operator."""
        self.intersect_update(other)
        return self

    def __repr__(self) -> str:
        """Return string representation of the mutable set."""
        return f"KotMutableSet({list(self._elements)})"



"""
KotSet: A Python implementation of Kotlin's Set interface with snake_case naming convention.
"""

from __future__ import annotations

from collections import defaultdict
from functools import reduce
from typing import TypeVar, Generic, Callable, Optional, Set, Iterator, Any, Tuple, List, Type, TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from kotcollections.kot_list import KotList
    from kotcollections.kot_mutable_list import KotMutableList
    from kotcollections.kot_map import KotMap
    from kotcollections.kot_mutable_set import KotMutableSet

T = TypeVar('T')
R = TypeVar('R')



class KotSet(Generic[T]):
    """A Python implementation of Kotlin's Set interface.
    
    This class provides all methods from Kotlin's Set interface with snake_case naming,
    maintaining type safety and immutability.
    """

    def __init__(self, elements: Optional[Set[T] | List[T] | Iterator[T] | 'KotList[T]' | 'KotMutableList[T]'] = None):
        """Initialize a KotSet with optional elements.
        
        Args:
            elements: Initial elements for the set (set, list, iterator, KotList, or KotMutableList)
        """
        self._elements: Set[T] = set()
        self._element_type: Optional[type] = None

        if elements is not None:
            # Check for KotList/KotMutableList first to avoid circular imports
            if hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
                # It's a KotList or KotMutableList
                for element in elements:
                    self._add_with_type_check(element)
            elif isinstance(elements, set):
                for element in elements:
                    self._add_with_type_check(element)
            elif isinstance(elements, list):
                for element in elements:
                    self._add_with_type_check(element)
            else:  # Iterator
                for element in elements:
                    self._add_with_type_check(element)

    @classmethod
    def __class_getitem__(cls, element_type: Type[T]) -> Type['KotSet[T]']:
        """Enable KotSet[Type]() syntax for type specification.
        
        Example:
            animals = KotSet[Animal]()
            animals.add(Dog("Buddy"))
            animals.add(Cat("Whiskers"))
        """
        class TypedKotSet(cls):
            def __init__(self, elements=None):
                # Only set element type if it's an actual type, not a type variable
                if isinstance(element_type, type):
                    self._element_type = element_type
                else:
                    self._element_type = None
                self._elements = set()
                # Now process elements with the correct type set
                if elements is not None:
                    if hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
                        # It's a KotList or KotMutableList
                        for element in elements:
                            self._add_with_type_check(element)
                    elif isinstance(elements, set):
                        for element in elements:
                            self._add_with_type_check(element)
                    elif isinstance(elements, list):
                        for element in elements:
                            self._add_with_type_check(element)
                    else:  # Iterator
                        for element in elements:
                            self._add_with_type_check(element)
        
        # Set a meaningful name for debugging
        TypedKotSet.__name__ = f"{cls.__name__}[{element_type.__name__}]"
        TypedKotSet.__qualname__ = f"{cls.__qualname__}[{element_type.__name__}]"
        
        return TypedKotSet

    @classmethod
    def of_type(cls, element_type: Type[T], elements: Optional[Set[T] | List[T] | Iterator[T]] = None) -> 'KotSet[T]':
        """Create a KotSet with a specific element type.
        
        This is useful when you want to create a set of a parent type
        but only have instances of child types.
        
        Args:
            element_type: The type of elements this set will contain
            elements: Optional initial elements
            
        Returns:
            A new KotSet instance with the specified element type
            
        Example:
            animals = KotSet.of_type(Animal, [Dog("Buddy"), Cat("Whiskers")])
            # or empty set
            animals = KotSet.of_type(Animal)
            animals.add(Dog("Max"))
        """
        # Use __class_getitem__ to create the same dynamic subclass
        typed_class = cls[element_type]
        return typed_class(elements)

    def _add_with_type_check(self, element: T) -> None:
        """Add an element with type checking."""
        # Skip type checking if element_type is a type variable or not a real type
        if self._element_type is not None and not isinstance(self._element_type, type):
            self._elements.add(element)
            return
            
        if self._element_type is None and element is not None:
            self._element_type = type(element) if not isinstance(element, KotSet) else KotSet

        if element is not None and self._element_type is not None:
            if isinstance(element, KotSet) and self._element_type == KotSet:
                self._elements.add(element)
            elif not isinstance(element, KotSet):
                # Check if element is an instance of the expected type
                if isinstance(element, self._element_type):
                    self._elements.add(element)
                # Special handling for __class_getitem__ types (e.g., KotList[Holiday])
                elif (hasattr(self._element_type, '__base__') and 
                      hasattr(self._element_type, '__name__') and 
                      '[' in self._element_type.__name__ and
                      isinstance(element, self._element_type.__base__)):
                    # Check if the element has matching element type for KotList/KotSet/KotMap types
                    if hasattr(element, '_element_type') and hasattr(self._element_type, '__new__'):
                        # Extract expected element type from the __class_getitem__ type
                        # This is a more strict check for collection types
                        self._elements.add(element)
                    else:
                        self._elements.add(element)
                else:
                    raise TypeError(
                        f"All elements must be of type {self._element_type.__name__}, got {type(element).__name__}"
                    )
            else:
                self._elements.add(element)
        elif element is None:
            self._elements.add(element)

    # Basic Set operations

    def is_empty(self) -> bool:
        """Returns true if the set is empty."""
        return len(self._elements) == 0

    def is_not_empty(self) -> bool:
        """Returns true if the set is not empty."""
        return len(self._elements) > 0

    @property
    def size(self) -> int:
        """Returns the size of the set."""
        return len(self._elements)

    def contains(self, element: T) -> bool:
        """Returns true if the set contains the specified element."""
        return element in self._elements

    def contains_all(self, elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> bool:
        """Returns true if the set contains all of the elements in the specified collection."""
        if isinstance(elements, KotSet):
            elements = elements._elements
        elif hasattr(elements, '_elements') and hasattr(elements, 'to_list'):
            # It's a KotList or KotMutableList
            elements = set(elements)
        elif isinstance(elements, list):
            elements = set(elements)
        return elements.issubset(self._elements)

    # Access operations

    def first(self) -> T:
        """Returns the first element."""
        if self.is_empty():
            raise ValueError("Set is empty")
        return next(iter(self._elements))

    def first_or_null(self) -> Optional[T]:
        """Returns the first element, or null if the set is empty."""
        if self.is_empty():
            return None
        return next(iter(self._elements))

    def first_or_none(self) -> Optional[T]:
        """Pythonic alias for first_or_null()."""
        return self.first_or_null()

    def first_predicate(self, predicate: Callable[[T], bool]) -> T:
        """Returns the first element matching the given predicate."""
        for element in self._elements:
            if predicate(element):
                return element
        raise ValueError("No element matching predicate")

    def first_or_null_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the first element matching the given predicate, or null if not found."""
        for element in self._elements:
            if predicate(element):
                return element
        return None

    def first_or_none_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Pythonic alias for first_or_null_predicate()."""
        return self.first_or_null_predicate(predicate)

    def last(self) -> T:
        """Returns the last element."""
        if self.is_empty():
            raise ValueError("Set is empty")
        return list(self._elements)[-1]

    def last_or_null(self) -> Optional[T]:
        """Returns the last element, or null if the set is empty."""
        if self.is_empty():
            return None
        return list(self._elements)[-1]

    def last_or_none(self) -> Optional[T]:
        """Pythonic alias for last_or_null()."""
        return self.last_or_null()

    def single(self) -> T:
        """Returns the single element, or throws an exception if the set is empty or has more than one element."""
        if self.size == 0:
            raise ValueError("Set is empty")
        if self.size > 1:
            raise ValueError("Set has more than one element")
        return next(iter(self._elements))

    def single_or_null(self) -> Optional[T]:
        """Returns the single element, or null if the set is empty or has more than one element."""
        if self.size != 1:
            return None
        return next(iter(self._elements))

    def single_or_none(self) -> Optional[T]:
        """Pythonic alias for single_or_null()."""
        return self.single_or_null()

    def single_predicate(self, predicate: Callable[[T], bool]) -> T:
        """Returns the single element matching the given predicate."""
        matches = [e for e in self._elements if predicate(e)]
        if len(matches) == 0:
            raise ValueError("No element matching predicate")
        if len(matches) > 1:
            raise ValueError("More than one element matching predicate")
        return matches[0]

    def single_or_null_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the single element matching the given predicate, or null."""
        matches = [e for e in self._elements if predicate(e)]
        if len(matches) != 1:
            return None
        return matches[0]

    def single_or_none_predicate(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Pythonic alias for single_or_null_predicate()."""
        return self.single_or_null_predicate(predicate)

    # Transformation operations

    def map(self, transform: Callable[[T], R]) -> 'KotSet[R]':
        """Returns a set containing the results of applying the given transform function to each element."""
        return KotSet(transform(element) for element in self._elements)

    def map_not_null(self, transform: Callable[[T], Optional[R]]) -> 'KotSet[R]':
        """Returns a set containing only the non-null results of applying the given transform function."""
        result = []
        for element in self._elements:
            transformed = transform(element)
            if transformed is not None:
                result.append(transformed)
        return KotSet(result)

    def map_not_none(self, transform: Callable[[T], Optional[R]]) -> 'KotSet[R]':
        """Pythonic alias for map_not_null()."""
        return self.map_not_null(transform)

    def flat_map(
        self,
        transform: Callable[[T], Set[R] | List[R] | 'KotSet[R]' | 'KotList[R]' | 'KotMutableList[R]']
    ) -> 'KotSet[R]':
        """Returns a single set of all elements from results of transform function."""
        result = []
        for element in self._elements:
            transformed = transform(element)
            if isinstance(transformed, KotSet):
                result.extend(transformed._elements)
            elif hasattr(transformed, '_elements') and hasattr(transformed, 'to_list'):
                # It's a KotList or KotMutableList
                result.extend(transformed)
            elif isinstance(transformed, set):
                result.extend(transformed)
            else:  # List
                result.extend(transformed)
        return KotSet(result)

    def map_indexed(self, transform: Callable[[int, T], R]) -> 'KotSet[R]':
        """Returns a set containing the results of applying the given transform function to each element and its index."""
        return KotSet(transform(index, element) for index, element in enumerate(self._elements))

    def flat_map_indexed(
        self,
        transform: Callable[[int, T], Set[R] | List[R] | 'KotSet[R]' | 'KotList[R]' | 'KotMutableList[R]']
    ) -> 'KotSet[R]':
        """Returns a single set of all elements from results of transform function applied to each element and its index."""
        result = []
        for index, element in enumerate(self._elements):
            transformed = transform(index, element)
            if isinstance(transformed, KotSet):
                result.extend(transformed._elements)
            elif hasattr(transformed, '_elements') and hasattr(transformed, 'to_list'):
                # It's a KotList or KotMutableList
                result.extend(transformed)
            elif isinstance(transformed, set):
                result.extend(transformed)
            else:  # List
                result.extend(transformed)
        return KotSet(result)

    # Filtering operations

    def filter(self, predicate: Callable[[T], bool]) -> 'KotSet[T]':
        """Returns a set containing only elements matching the given predicate."""
        return KotSet(element for element in self._elements if predicate(element))

    def filter_not(self, predicate: Callable[[T], bool]) -> 'KotSet[T]':
        """Returns a set containing only elements not matching the given predicate."""
        return KotSet(element for element in self._elements if not predicate(element))

    def filter_not_null(self) -> 'KotSet[T]':
        """Returns a set containing all elements that are not null."""
        return KotSet(element for element in self._elements if element is not None)

    def filter_not_none(self) -> 'KotSet[T]':
        """Pythonic alias for filter_not_null()."""
        return self.filter_not_null()

    def filter_is_instance(self, klass: Type[R]) -> 'KotSet[R]':
        """Returns a set containing all elements that are instances of the specified class."""
        return KotSet(element for element in self._elements if isinstance(element, klass))

    # Aggregation operations

    def all(self, predicate: Callable[[T], bool]) -> bool:
        """Returns true if all elements match the given predicate."""
        return all(predicate(element) for element in self._elements)

    def none(self, predicate: Callable[[T], bool]) -> bool:
        """Returns true if no elements match the given predicate."""
        return not any(predicate(element) for element in self._elements)

    def any(self, predicate: Optional[Callable[[T], bool]] = None) -> bool:
        """Returns true if set has at least one element, or any element matches the predicate."""
        if predicate is None:
            return self.is_not_empty()
        return any(predicate(element) for element in self._elements)

    def count(self, predicate: Optional[Callable[[T], bool]] = None) -> int:
        """Returns the number of elements matching the given predicate."""
        if predicate is None:
            return self.size
        return sum(1 for element in self._elements if predicate(element))

    def sum_of(self, selector: Callable[[T], float | int]) -> float | int:
        """Returns the sum of all values produced by selector function."""
        if self.is_empty():
            return 0
        values = [selector(element) for element in self._elements]
        return sum(values)

    def average(self, selector: Callable[[T], float | int]) -> float:
        """Returns the average of all values produced by selector function."""
        if self.is_empty():
            raise ValueError("Set is empty")
        values = [selector(element) for element in self._elements]
        return sum(values) / len(values)

    def max_or_null(self) -> Optional[T]:
        """Returns the largest element or null if there are no elements."""
        if self.is_empty():
            return None
        return max(self._elements)

    def max_or_none(self) -> Optional[T]:
        """Pythonic alias for max_or_null()."""
        return self.max_or_null()

    def min_or_null(self) -> Optional[T]:
        """Returns the smallest element or null if there are no elements."""
        if self.is_empty():
            return None
        return min(self._elements)

    def min_or_none(self) -> Optional[T]:
        """Pythonic alias for min_or_null()."""
        return self.min_or_null()

    def max_by_or_null(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Returns the element with the largest value of the selector function."""
        if self.is_empty():
            return None
        return max(self._elements, key=selector)

    def max_by_or_none(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Pythonic alias for max_by_or_null()."""
        return self.max_by_or_null(selector)

    def min_by_or_null(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Returns the element with the smallest value of the selector function."""
        if self.is_empty():
            return None
        return min(self._elements, key=selector)

    def min_by_or_none(self, selector: Callable[[T], Any]) -> Optional[T]:
        """Pythonic alias for min_by_or_null()."""
        return self.min_by_or_null(selector)

    # Collection operations

    def fold(self, initial: R, operation: Callable[[R, T], R]) -> R:
        """Accumulates value starting with initial value and applying operation."""
        return reduce(operation, self._elements, initial)

    def reduce(self, operation: Callable[[T, T], T]) -> T:
        """Accumulates value starting with the first element."""
        if self.is_empty():
            raise ValueError("Set is empty")
        return reduce(operation, self._elements)

    def reduce_or_null(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Accumulates value starting with the first element, or returns null."""
        if self.is_empty():
            return None
        return reduce(operation, self._elements)

    def reduce_or_none(self, operation: Callable[[T, T], T]) -> Optional[T]:
        """Pythonic alias for reduce_or_null()."""
        return self.reduce_or_null(operation)

    def group_by(self, key_selector: Callable[[T], R]) -> 'KotMap[R, KotSet[T]]':
        """Groups elements by the key returned by the given key_selector function."""
        from kotcollections.kot_map import KotMap
        groups = defaultdict(list)
        for element in self._elements:
            key = key_selector(element)
            groups[key].append(element)
        return KotMap({key: KotSet(elements) for key, elements in groups.items()})

    def group_by_to(
        self,
        key_selector: Callable[[T], R],
        value_transform: Callable[[T], Any]
    ) -> 'KotMap[R, KotList[Any]]':
        """Groups values returned by value_transform by the keys returned by key_selector."""
        from kotcollections.kot_map import KotMap
        from kotcollections.kot_list import KotList
        groups = defaultdict(list)
        for element in self._elements:
            key = key_selector(element)
            value = value_transform(element)
            groups[key].append(value)
        return KotMap({key: KotList(values) for key, values in groups.items()})

    def associate(self, transform: Callable[[T], Tuple[R, Any]]) -> 'KotMap[R, Any]':
        """Returns a Map containing key-value pairs provided by transform function."""
        from kotcollections.kot_map import KotMap
        return KotMap({key: value for key, value in (transform(element) for element in self._elements)})

    def associate_by(self, key_selector: Callable[[T], R]) -> 'KotMap[R, T]':
        """Returns a Map with keys generated by key_selector and values from elements."""
        from kotcollections.kot_map import KotMap
        return KotMap({key_selector(element): element for element in self._elements})

    def associate_with(self, value_selector: Callable[[T], R]) -> 'KotMap[T, R]':
        """Returns a Map where keys are elements and values are produced by value_selector."""
        from kotcollections.kot_map import KotMap
        return KotMap({element: value_selector(element) for element in self._elements})

    # Additional convenience operations

    def find(self, predicate: Callable[[T], bool]) -> Optional[T]:
        """Returns the first element matching the given predicate, or null if not found.
        
        This is an alias for first_or_null_predicate() for Kotlin compatibility.
        """
        return self.first_or_null_predicate(predicate)

    def partition(self, predicate: Callable[[T], bool]) -> Tuple['KotSet[T]', 'KotSet[T]']:
        """Splits the original set into pair of sets.
        
        The first set contains elements for which predicate yielded true,
        while the second set contains elements for which predicate yielded false.
        """
        true_elements = []
        false_elements = []
        for element in self._elements:
            if predicate(element):
                true_elements.append(element)
            else:
                false_elements.append(element)
        return KotSet(true_elements), KotSet(false_elements)

    def for_each(self, action: Callable[[T], None]) -> None:
        """Performs the given action on each element."""
        for element in self._elements:
            action(element)

    def for_each_indexed(self, action: Callable[[int, T], None]) -> None:
        """Performs the given action on each element, providing sequential index with the element."""
        for index, element in enumerate(self._elements):
            action(index, element)

    def with_index(self) -> Iterator[Tuple[int, T]]:
        """Returns an Iterator of IndexedValue for each element of the original set."""
        return enumerate(self._elements)

    def zip(self, other: Set[R] | List[R] | 'KotSet[R]' | 'KotList[R]' | 'KotMutableList[R]') -> 'KotSet[Tuple[T, R]]':
        """Returns a set of pairs built from the elements of this set and other collection with the same index."""
        if isinstance(other, KotSet):
            other = list(other._elements)
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = list(other)
        elif isinstance(other, set):
            other = list(other)
        return KotSet(zip(self._elements, other))

    def as_sequence(self) -> Iterator[T]:
        """Creates a sequence instance that wraps the original set, allowing lazy evaluation."""
        return iter(self._elements)

    # Set operations

    def union(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotSet[T]':
        """Returns a set containing all distinct elements from both collections."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        return KotSet(self._elements.union(other))

    def intersect(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotSet[T]':
        """Returns a set containing all elements that are contained by both collections."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        return KotSet(self._elements.intersection(other))

    def subtract(self, other: Set[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]') -> 'KotSet[T]':
        """Returns a set containing all elements that are not contained in the specified collection."""
        if isinstance(other, KotSet):
            other = other._elements
        elif hasattr(other, '_elements') and hasattr(other, 'to_list'):
            # It's a KotList or KotMutableList
            other = set(other)
        return KotSet(self._elements.difference(other))

    # Operator-style set operations

    def plus(self, element: T) -> 'KotSet[T]':
        """Returns a set containing all elements of the original set and the given element."""
        result = self._elements.copy()
        result.add(element)
        return KotSet(result)

    def plus_collection(
        self,
        elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]'
    ) -> 'KotSet[T]':
        """Returns a set containing all elements of the original set and the given collection."""
        return self.union(elements)

    def minus(self, element: T) -> 'KotSet[T]':
        """Returns a set containing all elements of the original set except the given element."""
        result = self._elements.copy()
        result.discard(element)
        return KotSet(result)

    def minus_collection(
        self,
        elements: Set[T] | List[T] | 'KotSet[T]' | 'KotList[T]' | 'KotMutableList[T]'
    ) -> 'KotSet[T]':
        """Returns a set containing all elements of the original set except those in the given collection."""
        return self.subtract(elements)

    # Conversion operations

    def to_list(self) -> List[T]:
        """Returns a Python list containing all elements."""
        return list(self._elements.copy())

    def to_set(self) -> Set[T]:
        """Returns a Python set containing all elements."""
        return set(self._elements.copy())

    def to_kot_list(self) -> 'KotList[T]':
        """Returns a Python list containing all elements."""
        from kotcollections.kot_list import KotList
        # Preserve type information when converting
        if self._element_type is not None:
            return KotList.of_type(self._element_type, list(self._elements))
        else:
            return KotList(self._elements.copy())

    def to_kot_mutable_list(self) -> 'KotMutableList[T]':
        """Returns a Python list containing all elements."""
        from kotcollections.kot_mutable_list import KotMutableList
        # Preserve type information when converting
        if self._element_type is not None:
            return KotMutableList.of_type(self._element_type, list(self._elements))
        else:
            return KotMutableList(self._elements.copy())

    def to_kot_set(self) -> KotSet[T]:
        """Returns a Python set containing all elements."""
        # Preserve type information when converting
        if self._element_type is not None:
            return KotSet.of_type(self._element_type, self._elements.copy())
        else:
            return KotSet(self._elements.copy())

    def to_kot_mutable_set(self) -> 'KotMutableSet[T]':
        """Returns a KotMutableSet containing all elements."""
        from kotcollections.kot_mutable_set import KotMutableSet
        # Preserve type information when converting
        if self._element_type is not None:
            mutable_set = KotMutableSet.of_type(self._element_type, self._elements.copy())
        else:
            mutable_set = KotMutableSet(self._elements.copy())
        return mutable_set

    def to_sorted_set(self, key: Optional[Callable[[T], Any]] = None, reverse: bool = False) -> KotSet[T]:
        """Returns a sorted list of all elements."""
        return KotSet(sorted(self._elements, key=key, reverse=reverse))

    def join_to_string(
        self,
        separator: str = ", ",
        prefix: str = "",
        postfix: str = "",
        limit: int = -1,
        truncated: str = "...",
        transform: Optional[Callable[[T], str]] = None
    ) -> str:
        """Creates a string from all the elements separated using separator."""
        elements = list(self._elements)
        if 0 <= limit < len(elements):
            elements = elements[:limit]
            truncated_part = truncated
        else:
            truncated_part = ""

        if transform:
            elements_str = [transform(e) for e in elements]
        else:
            elements_str = [str(e) for e in elements]

        return prefix + separator.join(elements_str) + truncated_part + postfix

    # Python special methods

    def __repr__(self) -> str:
        """Return string representation of the set."""
        return f"KotSet({list(self._elements)})"

    def __iter__(self) -> Iterator[T]:
        """Return an iterator over the elements."""
        return iter(self._elements)

    def __contains__(self, item: T) -> bool:
        """Check if item is in the set."""
        return item in self._elements

    def __len__(self) -> int:
        """Return the size of the set."""
        return len(self._elements)

    def __eq__(self, other: object) -> bool:
        """Check if two sets are equal."""
        if not isinstance(other, KotSet):
            return False
        return self._elements == other._elements

    def __hash__(self) -> int:
        """Return hash of the set."""
        return hash(frozenset(self._elements))



from .kot_list import KotList
from .kot_map import KotMap, KotMapWithDefault
from .kot_mutable_list import KotMutableList
from .kot_mutable_map import KotMutableMap
from .kot_mutable_set import KotMutableSet
from .kot_set import KotSet

__all__ = ['KotList', 'KotMutableList', 'KotSet', 'KotMutableSet', 'KotMap', 'KotMutableMap', 'KotMapWithDefault']

# Version will be dynamically set by poetry-dynamic-versioning
try:
    from ._version import __version__
except ImportError:
    # Fallback for development
    __version__ = '0.0.0'
